import datetime
import sys
import time
from wiliot_deployment_tools.api.extended_api import BridgeThroughGatewayAction, ExtendedEdgeClient, GatewayType,ExtendedPlatformClient,EXCEPTIONS_TO_CATCH,BoardTypes
from wiliot_deployment_tools.common.debug import debug_print
from wiliot_deployment_tools.common.utils_defines import  SEC_TO_SEND,BROADCAST_DST_MAC


class GwHbActionClientError(Exception):
    pass
  
       
class GwHbActionClient(ExtendedEdgeClient):

    def __init__(self, edge_api_key, asset_api_key, owner,env='prod', cloud="", region="us-east-2"):
        # Initialize objects
        self.edge = ExtendedEdgeClient(edge_api_key, owner, env, cloud=cloud, region=region)
        self.plat = ExtendedPlatformClient(asset_api_key, owner, env, cloud=cloud, region=region)


    def get_relevant_bridges(self, gateway_id):
        """
        Returns seen bridges by gateway that are associated to owner and have ble version with gw heat beat support
        :type gateway_id: str
        :param gateway_id: gateway ID
        :rtype: list
        :return: list of relevant bridge IDs
        """
        bridges = self.edge.get_seen_bridges(gateway_id)
        relevant_bridges = bridges
        for bridge in bridges:
            brg_ble = self.edge.get_brg_ble_version(bridge)
            if not self.check_ble_gw_hb_support(brg_ble):
                # debug_print(f'Bridge {bridge} does not support gw heat beat!')
                relevant_bridges.remove(bridge)
        return relevant_bridges


    @staticmethod
    def check_ble_gw_hb_support(ble_ver):
        """
        checks if gw_hb management is supported for BLE version
        :rtype: bool
        :return: True if supported, False otherwise
        """
        ble_ver = ble_ver.split('.')
        if not ((int(ble_ver[0]) > 3) or
                (int(ble_ver[0]) > 2 and int(ble_ver[1]) > 15) or \
                (int(ble_ver[0]) > 2 and int(ble_ver[1]) > 14 and int(ble_ver[2]) > 28)):
            return False
        else:
            return True
    
    
    def send_gw_hb_pkt(self, gateway_id,payload,silent=False, seconds_to_send=1):
        """
        send gw heat beat packet to bridge
        :type bridge_id: str
        :param bridge_id: Bridge ID
        :type packet: GwHbActionClientPacket
        :param packet: packet to send to bridge
        :type gateway_id: str
        :param gateway_id: Gateway ID
        :type silent: bool
        :param silent: print gw heat beat packet parameters
        :type seconds_to_send: int
        :param seconds_to_send: seconds to repeat packet
        :rtype: bool
        :return: True if sent, False otherwise
        """
        
        bridge_id = BROADCAST_DST_MAC
        via_gw = gateway_id
        reps = 8
        seconds_to_send = datetime.timedelta(seconds=seconds_to_send)
        # Check BRG Compatability
        brg_version = None
        # Check GW Compatability
        gw_type = self.edge.get_gateway_type(via_gw)
        if gw_type == GatewayType.WIFI:
            gw_version = self.edge.get_gw_ble_version(via_gw)
            rxtx = datetime.timedelta(milliseconds=
                                    self.edge.get_gateway(via_gw)['reportedConf']['additional']['rxTxPeriodMs'])
            if not silent:
                debug_print(f'Sending packet through WIFI GW {via_gw}')
        if gw_type == GatewayType.MOBILE:
            if not silent:
                debug_print(f'Sending packet through ANDROID GW {via_gw}')
            rxtx = datetime.timedelta(milliseconds=140)
        reps = seconds_to_send // rxtx + 1
        try:
            res = self.edge.send_bridge_action_through_gw(gateway_id,
                                                 BridgeThroughGatewayAction.GW_HB,
                                                 payload, bridge_id, reps=reps)
        except EXCEPTIONS_TO_CATCH as e:
            debug_print('Exception caught when sending gw hb packet!')
            debug_print(e)
            return False
        return res

    
    def check_gw_hb_ack_received(self, gateway_id,payload, time_sent):
        """
        check if gw heat beat packet was received by bridge
        :type gateway_id: str
        :param gateway_id: Gateway ID
        :type payload: str
        :param payload containing gw_id
        :type time_sent: datetime
        :param time_sent: time packet was sent
        :rtype dict
        :return: dictionary with brg_id as key, rssi as value
        """
        hours_back = (datetime.datetime.now() - time_sent + datetime.timedelta(seconds=5)).total_seconds() / 3600
        acks = self.edge.get_acks(gateway_id, hours_back)
        rssi_dict = {}
        if acks is None:
            return None
        ack_ready = False
        for line in range(len(acks)):
            ack_payload  = acks['payload'].iloc[line]
            ack_gw_id = ack_payload[0:11]
            if  ack_gw_id == payload[:12] and acks['actionType'].iloc[line] == BridgeThroughGatewayAction.GW_HB:
                ack_ready = True
                brg_mac = acks['bridgeId'].iloc[line]
                brg_ble = self.edge.get_brg_ble_version(brg_mac)
                if self.edge.check_brg_mel_mod_supprt(brg_mac):
                    rssi = int(ack_payload[-2:], 16)
                else:
                    rssi = int(ack_payload[12:14], 16)
                if brg_mac in rssi_dict:
                    # update rssi to max vaule received
                    if rssi > rssi_dict[brg_mac]:
                        rssi_dict[brg_mac] = rssi 
                else:
                    rssi_dict[brg_mac] = rssi  
        if ack_ready:
            return rssi_dict
        return None
   

    def send_gw_hb_packet_to_brgs_list(self,bridges,gateways,minutes_timeout=3,rssi_th = 76,brg_gw_dict = None,is_bridge_log = False):
        """
        send gw hb from all gateways and remove bridgese with good rssi
        :type bridges: list
        :param bridges: Bridges IDs list
        :type gateways: list
        :param gateways: Gateways IDs list
        :type minutes_timeout: int
        :param minutes_timeout: timeout
        :type rssi_th int
        :param rssi value threshold for OTA
        :return: List of bridges with low connection
        """
        gws_dict = {}
        gws_rssi_dict = {} #for bridge log option
        time_sent = datetime.datetime.now()
        for gw in gateways:
            payload = gw[2:13] 
            gws_dict[gw] = payload
        debug_print(f'Broadcasting GW HB throgh gws: {gateways} | Max Waiting time {minutes_timeout} minutes')
        # broadcast gw_hb through all gateways
        for gw in gateways:
            self.send_gw_hb_pkt(gw,gws_dict[gw], silent=False, seconds_to_send=SEC_TO_SEND)
        time_gw_stops = time_sent + datetime.timedelta(seconds=SEC_TO_SEND)
        time_end = time_sent + datetime.timedelta(minutes=minutes_timeout)
        curr_time = time_sent
        while curr_time < time_end and bridges != []:
            curr_time = datetime.datetime.now()
            if curr_time > time_gw_stops - datetime.timedelta(seconds=1):  # 1 second buffer
                # broadcast gw_hb through all gateways
                for gw in gateways:
                    self.send_gw_hb_pkt(gw,gws_dict[gw], silent=False, seconds_to_send=SEC_TO_SEND)
                time_gw_stops = datetime.datetime.now() + datetime.timedelta(seconds=SEC_TO_SEND)
            time.sleep(0.5)
            for gw in gateways:
                # check for acks of all gws
                rssi_dict = self.check_gw_hb_ack_received(gw,gws_dict[gw], time_sent)
                if rssi_dict: # key: brg value: rssi
                    if is_bridge_log:
                                gws_rssi_dict[gw] = rssi_dict[bridges[0]]
                                gateways.remove(gw)
                                continue
                    for brg_id in bridges:
                        if brg_id in rssi_dict and brg_id in brg_gw_dict and gw in brg_gw_dict[brg_id]:
                            gw_info = brg_gw_dict[brg_id][gw] # gws dict
                            # gw_info = gws_dict_for_brg[gw]
                            if rssi_dict[brg_id] > gw_info[1]: # update min rssi
                                gw_info[1] = rssi_dict[brg_id]
                            if gw_info[0] > 1 and curr_time < time_end: #check msg cnt
                                # in case of timeout, take current min rssi value
                                gw_info[0] = gw_info[0] - 1
                                continue
                            else: #check rssi th                                      
                                if gw_info[1] <= rssi_th and not is_bridge_log:
                                    # remove bridges with good connection 
                                    debug_print(f'got good connection for bridge:{brg_id} with rssi:{gw_info[1]}')
                                    bridges.remove(brg_id)
                                    if bridges != []:
                                        debug_print(f'continue to next bridges:{bridges}')
            if not gateways and is_bridge_log:
                return gws_rssi_dict
        return bridges
        
        

   
    def check_bridges_for_gw_hb_action(self,bridges):
        if not bridges:
            return None
        for bridge in bridges:
            try:
                brg_version = self.edge.get_brg_ble_version(bridge)
            except Exception as e:
                if "not owned" in str(e):
                    bridges.remove(bridge)
                    debug_print(f"Warning!! bridge {bridge} is not owned")
                    continue
            brg_status = self.edge.get_bridge_status(bridge)
            if brg_status == 'offline':
                bridges.remove(bridge)
                debug_print(f"Warning: bridge {bridge} is offline.")
                continue
            if not self.check_ble_gw_hb_support(brg_version):
                bridges.remove(bridge)
                debug_print(f"Warning: bridge {bridge} version doesn't support the check. Please update version.")
        return bridges

    def create_brg_gw_cnt_dict(self, bridges):
        '''
        create a dictionary, with brg_id as key
        value: a dictionary of gateways seeing this bridge (as a key)
        value: a list of counter and min rssi
        example for dictionary:
        brg_gw_cnt_dict[brg_x] -> gws_cnt_dict[gw] = [10,0]

        '''
        minew_rssi_msgs_cnt = 10  # count rssi packets before checking threshold
        bridge_rssi_msgs_cnt = 1
        min_rssi = 0
        brg_gw_cnt_dict = {}
        for brg in bridges:
            gws_of_brg = list(self.edge.get_gateways_from_bridges(brg))
            gws_cnt_dict = {}
            for gw in gws_of_brg:
                if self.edge.check_gw_compatible_for_action(gw):
                    if self.edge.get_bridge_board(brg) == BoardTypes.MINEW_DUAL.value:
                        gws_cnt_dict[gw] = [minew_rssi_msgs_cnt,min_rssi]
                    else:
                        gws_cnt_dict[gw] = [bridge_rssi_msgs_cnt,min_rssi]
            #add new tuple to current bridge list
            brg_gw_cnt_dict[brg] = gws_cnt_dict
        return brg_gw_cnt_dict

    def bridge_gw_downlink_analyzer(self, location=None, minutes_timeout=3,bridges_list = None,gws_list = None,is_bridge_log = False):
        """
        Get check rssi for OTA of all bridges by a given location
        :type location: str
        :param location: Location name to check (as written in platform)
        :type packet: PowerManagementPacket
        :param packet: packet to send
        :type minutes_timeout: int
        :param minutes_timeout: timeout
        :rtype: dict
        :return: dict of {bridgeID: ack status (bool)} pairs
        """
        rssi_th = 76
        bridges = []
        gateways = []
        # get all bridges in the given location/from cmd
        if bridges_list:
            bridges = bridges_list
            debug_print('got bridges list from cmd')
        elif location:
            debug_print(f'Getting bridges from {location}')
            bridges = self.plat.get_location_bridges(location)
        elif gws_list:
            debug_print(f'Getting bridges from given gateways')
            for gw in gws_list:
                gw_brgs = self.get_relevant_bridges(gw)
                for brg in gw_brgs:
                    if brg not in bridges:
                        bridges.append(brg)
        else:
            debug_print('Error!: must suplly Location, bridges or getways')
            return False   
                     
        bridges = self.check_bridges_for_gw_hb_action(bridges)

        if not bridges:
            debug_print('Error: no bridges found for analyzer tool')
            return False
        
        # get all gw that see thoes bridges
        if gws_list is not None:
            gateways = gws_list
            debug_print('got gateways list from cmd')
        else:
             for bridge in bridges:
                gws_of_brg = list(self.edge.get_gateways_from_bridges(bridge))
                for gw in gws_of_brg:
                    if gw not in gateways:
                        gateways.append(gw)
        for gw in gateways:
            if not self.edge.check_gw_compatible_for_action(gw):
                debug_print(f'Warning! gateway {gw} does not support Action!')
                gateways.remove(gw)
        if not gateways:
            debug_print('Error! no any valid gateways for check!')
            return False
        # in case of minew dual band bridge for each bridge,each gw
        #  holds a counter of rssi packet received to make sure receiving rssi of both antenas
        brg_gw_dict = self.create_brg_gw_cnt_dict(bridges)
        res = self.send_gw_hb_packet_to_brgs_list(bridges = bridges,minutes_timeout=minutes_timeout
                                                                      ,gateways=gateways,rssi_th=rssi_th,brg_gw_dict=brg_gw_dict,is_bridge_log=is_bridge_log)
        if not is_bridge_log:
            unconnected_bridges = res
            if unconnected_bridges:
                debug_print(f'Not all bridges have stable connection, check following bridge:{unconnected_bridges}')
                return False, res
            else:
                debug_print('Success! All bridges have stable downlink')
                return True, res
        else:
            gws_rssi_list = res
            debug_print(f'Printing bridge rssi log for bridge:{bridges_list}')
            for gateway, rssi in gws_rssi_list.items():
                debug_print(f"gateway: {gateway} with rssi {rssi}")



from argparse import ArgumentParser
import sys

from wiliot_core import check_user_config_is_ok
from wiliot_deployment_tools.firmware_update.firmware_update import FirmwareUpdate
from wiliot_deployment_tools.common.debug import debug_print
def main():
    parser = ArgumentParser(prog='wlt-firmware',
                            description='Firmware Update - CLI Tool for updating Wiliot Gateways and Bridges firmware OTA')
    required = parser.add_argument_group('required arguments')
    required.add_argument('-owner', type=str, help="Owner ID", required=True)
    parser.add_argument('-test', action='store_true',
                        help='If flag used, use test environment (prod is used by default)')
    parser.add_argument('-beta', action='store_true', help="Show available beta versions / update to beta firmware",
                               required=False)
    subparsers = parser.add_subparsers()

    parser_versions = subparsers.add_parser('versions', help='Display available versions',
                                            description='Prints all avaliable Firmware versions for update for every GW Type',
                                            epilog="example usage: wlt-firmware -o wiliot versions")
    parser_versions.set_defaults(func=print_versions)

    parser_update = subparsers.add_parser('update', help='Update Gateways and Bridges',
                                          description=' Run OTA Process, first updating specified GWs to latest / specified FW version. Afterwards seuqentially update each specified Bridges / all Bridges to the same Firmware version.',
                                          epilog="example usage: wlt-firmware -o wiliot update -gw GW0123 -all_brgs")
    parser_update.set_defaults(func=firmware_update)
    parser_update.add_argument('-version', type=str,
                               help="Desired version. If not specified, will update to latest available version",
                               required=False)
    parser_update.add_argument('-gw', nargs='+', help="Gateways to update (multiple allowed)", required=False)
    parser_update.add_argument('-brg', nargs='+', help="Bridges to update (multiple allowed)", required=False)
    parser_update.add_argument('-all_brgs', action='store_true', help="Update all bridges connected to Gateways", required=False)
    parser_update.add_argument('-ignore_bridges', nargs='+', required=False, help='Bridges to ignore')
    parser_update.add_argument('-action', action='store_true', help="Update using action API",
                               required=False)
    # parser_update.add_argument('-force', action='store_true',
    #                            help='update bridge even if its already in desired version (applicable only with -action)')

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)
        
    args = parser.parse_args()
    if args.test:
        env = 'test'
    else:
        env = 'prod'
    owner_id = args.owner
    conf_env = env if env == 'prod' else 'non-prod'
    user_config_file_path, api_key, is_success = check_user_config_is_ok(owner_id, conf_env, 'edge')
    if is_success:
        print('credentials saved/upload from {}'.format(user_config_file_path))
    else:
        raise Exception('invalid credentials - please try again to login')

    args.fw = FirmwareUpdate(api_key, owner_id, env)
    if 'func' not in args.__dict__:
        debug_print('Wrong Usage! please specify positional arguments')
        parser.print_help()
        sys.exit()
    args.func(args)


def firmware_update(args):
    # get specified GW types
    args.latest = True if args.version is None else False
    args.fw.firmware_update(gws_list=args.gw,
                             brgs_to_update=args.brg,
                             update_all_connected_bridges=args.all_brgs,
                             desired_version=args.version,
                             update_to_latest=args.latest,
                             include_beta=args.beta,
                             ignore_bridges=args.ignore_bridges,
                             action=args.action)
                            #  force_update=args.force)

def print_versions(args):
    args.fw.print_available_versions(args.beta)

def main_cli():
    main()

if __name__ == "__main__":
    main()



from wiliot_deployment_tools.interface.uart_if import UARTInterface
from wiliot_deployment_tools.common.debug import debug_print
from wiliot_deployment_tools.interface.if_defines import *
import datetime
import time

class BLESimulator():
    def __init__(self, uart:UARTInterface):
        self.uart = uart
        self.sim_mode = False

    def set_sim_mode(self, sim_mode):
        self.uart.flush()
        mode_dict = {True: 1, False: 0}
        self.sim_mode = sim_mode
        self.uart.reset_gw()
        self.uart.write_ble_command(f"!ble_sim_init {mode_dict[sim_mode]}")
        if not sim_mode:
            self.uart.cancel()
        time.sleep(3)

    def send_packet(self, raw_packet, duplicates=DEFAULT_DUPLICATES, output_power=DEFAULT_OUTPUT_POWER, channel=SEND_ALL_ADV_CHANNELS, delay=DEFAULT_DELAY):
        assert self.sim_mode is True, 'BLE Sim not initialized!'
        if len(raw_packet) == 62:
            # Add ADVA
            raw_packet = DEFAULT_ADVA + raw_packet
        if len(raw_packet) != 74:
            raise ValueError('Raw Packet must be 62/74 chars long!')
        self.uart.write_ble_command(f"!ble_sim {str(raw_packet)} {str(duplicates)} {str(output_power)} {str(channel)} {str(delay)}")
        if delay > 0:
            diff = time.perf_counter()
            time.sleep((delay/1000) * duplicates)
            diff = time.perf_counter() - diff
            debug_print(f'Desired Delay: {(delay/1000) * duplicates} Actual Delay {diff}')

    
    def send_data_si_pair(self, data_packet, si_packet, duplicates, output_power=DEFAULT_OUTPUT_POWER, delay=DEFAULT_DELAY, packet_error=None):
        if packet_error is None:
            packet_error = [True for i in range (duplicates * 2)]
        # debug_print(packet_error)
        # print(f'delay {delay}')
        packet_to_send = data_packet
        def switch_packet(packet_to_send):
            if packet_to_send == data_packet:
                return si_packet
            else:
                return data_packet
        for dup in range(duplicates * 2):
            diff = time.perf_counter()
            if packet_error[dup]:
                debug_print(f'Sending Packet {dup}')
                self.send_packet(packet_to_send, duplicates=1, output_power=output_power, channel=SEND_ALL_ADV_CHANNELS, delay=0)
            else:
                debug_print(f'Dropping Packet {dup}')
            time.sleep(delay/1000)
            diff = time.perf_counter() - diff
            debug_print(f'Desired Delay: {delay/1000} Actual Delay {diff}')
            packet_to_send = switch_packet(packet_to_send)
    
    def trigger_by_time_stamp(self, ts):
        if ts == None:
            return
        current_time = datetime.datetime.timestamp(datetime.datetime.now()) * 1000
        time_difference = ts-current_time
        print(f"The test will start in: {time_difference/1000} secondes")
        time.sleep(time_difference/1000)

    def send_brg_network_pkts(self, pkts, duplicates, output_power=DEFAULT_OUTPUT_POWER):
        num_brgs = len(pkts)
        total_pkts_to_send = num_brgs * 2 * duplicates
        for pkt_idx in range(total_pkts_to_send):
            brg_idx = pkt_idx % 3
            pkt = pkts[brg_idx]
            pkt_idx_per_brg = pkt_idx // num_brgs
            if not bool(pkt_idx % 2):
                packet_to_send = pkt['data_packet']
            else:
                packet_to_send = pkt['si_packet']
            packet_error = pkt['packet_error']
            if packet_error[pkt_idx_per_brg]:
                debug_print(f'BRG {pkt["bridge_id"]}: Sending Packet {pkt_idx_per_brg}')
                self.send_packet(packet_to_send, duplicates=1, output_power=output_power,
                                 channel=SEND_ALL_ADV_CHANNELS, delay=0)
            else:
                debug_print(f'BRG {pkt["bridge_id"]}: Dropping Packet {pkt_idx_per_brg}')
            diff = time.perf_counter()
            delay = pkt['time_delay']
            time.sleep(delay/1000)
            diff = time.perf_counter() - diff
            debug_print(f'Desired Delay: {delay/1000} Actual Delay {diff}')


from wiliot_deployment_tools.interface.if_defines import *
from wiliot_deployment_tools.ag.ut_defines import *

RECEIVED = 'received'
SHARED_COLUMNS = [PAYLOAD, BRIDGE_ID, NFPKT, RSSI]
INT64_COLUMNS = [NFPKT, RSSI]
OBJECT_COLUMNS = [PAYLOAD, BRIDGE_ID]
INIT_STAGES_DUPLICATIONS = [i for i in range(3,7)]
REPORT_COLUMNS = ['pkt_id', 'bridgeId', 'duplication', 'time_delay']
INCREMENTAL_TIME_DELAYS = [10, 50, 100, 255]
INCREMENTAL_STAGE_ADVA = '0000000000C0'
INCREMENTAL_PACKETS = range(255)


import datetime
import sys
import time
from wiliot_deployment_tools.api.extended_api import EXCEPTIONS_TO_CATCH, BridgeThroughGatewayAction, ExtendedEdgeClient, GatewayType
from wiliot_deployment_tools.common.debug import debug_print
from wiliot_deployment_tools.common.utils_defines import  SEC_TO_SEND, BROADCAST_DST_MAC


class CalibrationManagementError(Exception):
    pass
  

class CalibrationManagementClient(ExtendedEdgeClient):
    
    is_first_call = True

    def get_relevant_bridges(self, gateway_id):
        """
        Returns seen bridges by gateway that are associated to owner and have ble version with calibration management support
        :type gateway_id: str
        :param gateway_id: gateway ID
        :rtype: list
        :return: list of relevant bridge IDs
        """
        bridges = self.get_seen_bridges(gateway_id)
        relevant_bridges = bridges
        not_owned_brgs = []
        not_support_version = []
        for bridge in bridges:
            try:
                brg_ble = self.get_brg_ble_version(bridge)
            except Exception as e:
                if "not owned" in str(e):
                    not_owned_brgs.append(bridge)
                    relevant_bridges.remove(bridge)
                    continue
            if not self.check_ble_calibration_mgmt_support(brg_ble):
                not_support_version.append(bridge)
                relevant_bridges.remove(bridge)
        if self.is_first_call:
            # print warning once
            self.is_first_call = False
            debug_print(f"Warning! bridges id:{not_owned_brgs} are not owned, skiping bridges")
            debug_print(f'Bridges {not_support_version} do not support calibration management!')
        return relevant_bridges


    @staticmethod
    def check_ble_calibration_mgmt_support(ble_ver):
        """
        checks if calibration management is supported for BLE version
        :rtype: bool
        :return: True if supported, False otherwise
        """
        ble_ver = ble_ver.split('.')
        if not ((int(ble_ver[0]) > 3) or
                (int(ble_ver[0]) > 2 and int(ble_ver[1]) > 14) or \
                (int(ble_ver[0]) > 2 and int(ble_ver[1]) > 13 and int(ble_ver[2]) > 61)):
            return False
        else:
            return True


    def send_calib_mgmt_pkt(self, bridge_id, payload, gateway_id=None,
                          silent=False, seconds_to_send=1):
        """
        send calibration management packet to bridge
        :type bridge_id: str
        :param bridge_id: Bridge ID
        :type packet: CalibrationManagementPacket
        :param packet: packet to send to bridge
        :type gateway_id: str
        :param gateway_id: Gateway ID
        :type silent: bool
        :param silent: print calibration management packet parameters
        :type seconds_to_send: int
        :param seconds_to_send: seconds to repeat packet
        :rtype: bool
        :return: True if sent, False otherwise
        """
        via_gw = []
        if gateway_id is None:
            via_gw = self.get_bridge_relevant_gw(bridge_id, get_gw_list=True)
        else:
            via_gw.append(gateway_id)
        seconds_to_send = datetime.timedelta(seconds=seconds_to_send)
        for gw in via_gw:
            # Check BRG Compatability
            gw_type = self.get_gateway_type(gw)
            if gw_type == GatewayType.WIFI:
                rxtx = datetime.timedelta(milliseconds=
                                    self.get_gateway(gw)['reportedConf']['additional']['rxTxPeriodMs'])
            if gw_type == GatewayType.MOBILE:
                rxtx = datetime.timedelta(milliseconds=140)
            reps = seconds_to_send // rxtx + 1
            if not silent:
                debug_print(f'Sending packet through GW {gw} for {seconds_to_send.total_seconds()} seconds, rxTx = {int(rxtx.total_seconds()*1000)} ms, reps = {reps}')
            try:
                res = self.send_bridge_action_through_gw(gw,
                                                        BridgeThroughGatewayAction.SPARSE_37,
                                                        payload, bridge_id, reps=reps)
            except EXCEPTIONS_TO_CATCH as e:
                debug_print('Exception caught when sending calibration management packet!')
                debug_print(e)
                return False
        return res
    
    def check_calib_mgmt_received(self, gateway_id, bridge_id, payload, time_sent):
        """
        check if calibration management packet was received by bridge
        :type gateway_id: str
        :param gateway_id: Gateway ID
        :type bridge_id: str
        :param bridge_id: Bridge ID
        :type packet: CalibrationManagementPacket
        :param packet: packet to check
        :type time_sent: datetime
        :param time_sent: time packet was sent
        :rtype bool
        :return: True if packet was received, None otherwise
        """
        hours_back = (datetime.datetime.now() - time_sent + datetime.timedelta(seconds=5)).total_seconds() / 3600
        acks = self.get_acks(gateway_id, hours_back)
        if acks is None:
            return None
        acks = acks[(acks['actionType'] == BridgeThroughGatewayAction.SPARSE_37) &
                    (acks['bridgeId'] == bridge_id) &
                    (acks['payload'].apply(lambda x: x[1])
                     == payload[1])] 
        if len(acks) > 0:
            return acks.to_dict()
        return None

        
    def send_calibration_packet_until_ack(self, gateway_id, bridge_id, minutes_timeout=3, mode=0):
        """
        send calibration management packet repeatedly until it is received by the bridge
        :type gateway_id: str
        :param gateway_id: Gateway ID
        :type bridge_id: str
        :param bridge_id: Bridge ID
        :type packet: CalibrationManagementPacket
        :param packet: packet to send
        :type minutes_timeout: int
        :param minutes_timeout: timeout
        :rtype: dict
        :return: dict of {bridgeID: ack status (bool)} pairs
        """
        brg_version = self.get_brg_ble_version(bridge_id)
        if not self.check_ble_calibration_mgmt_support(brg_version):
            debug_print(f"Error! bridge:{bridge_id} ble version does'nt support calibration mgmt")
            return False
        time_sent = datetime.datetime.now()
        payload = '0{:01d}{:019d}'.format(mode, 0)
        debug_print(f'Configuring {bridge_id} To Calibration mode {mode} | Max Waiting time {minutes_timeout} minutes')
        self.send_calib_mgmt_pkt(bridge_id, payload, gateway_id, silent=False, seconds_to_send=SEC_TO_SEND)
        time_gw_stops = time_sent + datetime.timedelta(seconds=SEC_TO_SEND)
        time_end = time_sent + datetime.timedelta(minutes=minutes_timeout)
        curr_time = time_sent
        ack = None
        print('Waiting For Ack...', end='', flush=True)
        while curr_time < time_end and ack is None:
            curr_time = datetime.datetime.now()
            if curr_time > time_gw_stops - datetime.timedelta(seconds=1):  # 1 second buffer
                self.send_calib_mgmt_pkt(bridge_id, payload, gateway_id, silent=True, seconds_to_send=SEC_TO_SEND)
                time_gw_stops = datetime.datetime.now() + datetime.timedelta(seconds=SEC_TO_SEND)
                print('!', end='')
                sys.stdout.flush()
            time.sleep(0.5)
            print('.', end='')
            sys.stdout.flush()
            if gateway_id is None:
                gws = self.get_bridge_relevant_gw(bridge_id, get_gw_list=True)
                for gw in gws:
                    ack = self.check_calib_mgmt_received(gw, bridge_id, payload, time_sent)
                    if ack:
                        break
            else:
                ack = self.check_calib_mgmt_received(gateway_id, bridge_id, payload, time_sent)
        print()  # enter newline
        ack_received = False
        if ack is not None:
            ack_received = True
            debug_print(f'Ack Received after {datetime.datetime.now() - time_sent}')
        else:
            debug_print(f'Ack not received after {datetime.datetime.now() - time_sent}!')
        return {bridge_id: ack_received}
    

    def broadcast_calib_mgmt_packet(self, payload, gateway_id, silent=False, seconds_to_send=1):
        """
        broadcast packet to all bridges connected to GW
        :type packet: PowerManagementPacket
        :param packet: packet to broadcast
        :type gateway_id: str
        :param gateway_id: Gateway ID
        :type silent: bool
        :param silent: print power management packet parameters
        :rtype: tuple
        :return: results (bool), bridges sent to (list)
        """
        if not self.check_gw_compatible_for_action(gateway_id):
            raise CalibrationManagementError(f'Gateway {gateway_id} not compatible with power management! '
                                       f'Try to update GW version') 
        brgs_to_broadcast = self.get_relevant_bridges(gateway_id)
        for bridge in brgs_to_broadcast:
            try:
                bridge_ver = self.get_brg_ble_version(bridge)
            except Exception as e:
                if "not owned" in str(e):
                    brgs_to_broadcast.remove(bridge)
                    continue
            bridge_ver = self.get_brg_ble_version(bridge)
            if not self.check_ble_calibration_mgmt_support(bridge_ver):
                brgs_to_broadcast.remove(bridge)
        if not silent:
            debug_print(f'Broadcasting Packet to bridges {brgs_to_broadcast}')
        res = self.send_calib_mgmt_pkt(BROADCAST_DST_MAC, payload, gateway_id, silent, seconds_to_send)
        return res, brgs_to_broadcast


    def check_calib_mgmt_broadcast_received(self, gateway_id, bridges, payload, time_sent):
        """
        check if power management broadcast was received by bridges
        :type gateway_id: str
        :param gateway_id: Gateway ID
        :type bridges: list
        :param bridges: list of Bridge IDs
        :type time_sent: datetime
        :param time_sent: time packet was sent
        :rtype: dict
        :return: dict of bridgeIDs:received status
        """
        hours_back = (datetime.datetime.now() - time_sent + datetime.timedelta(seconds=10)).total_seconds() / 3600
        acks = self.get_acks(gateway_id, hours_back)
        result = dict().fromkeys(bridges)
        for brg in bridges:
            if acks is None:
                result[brg] = False
                continue
            brg_acks = acks[(acks['actionType'] == BridgeThroughGatewayAction.SPARSE_37) &
                    (acks['bridgeId'] == brg) &
                    (acks['payload'].apply(lambda x: x[1])
                     == payload[1])] 
            if len(brg_acks) > 0:
                result[brg] = True
            else:
                result[brg] = False
        return result


    def broadcast_calibration_packet_until_ack(self, gateway_id,minutes_timeout=3, mode=0):
        """
        send power management packet repeatedly until it is received by the all 'seen' bridges
        :type gateway_id: str
        :param gateway_id: Gateway ID
        :type packet: PowerManagementPacket
        :param packet: packet to send
        :type minutes_timeout: int
        :param minutes_timeout: timeout
        :rtype: dict
        :return: dict of {bridgeID: ack status (bool)} pairs
        """
        time_sent = datetime.datetime.now()
        payload = '0{:01d}{:019d}'.format(mode, 0)
        res, bridges = self.broadcast_calib_mgmt_packet(payload, gateway_id, silent=False, seconds_to_send=SEC_TO_SEND)
        debug_print(f'Configuring {bridges} To calibration mode {mode} | Max Waiting time {minutes_timeout} minutes')
        time_gw_stops = time_sent + datetime.timedelta(seconds=SEC_TO_SEND)
        time_end = time_sent + datetime.timedelta(minutes=minutes_timeout)
        curr_time = time_sent
        ack = self.check_calib_mgmt_broadcast_received(gateway_id, bridges, payload, time_sent)
        print('Waiting For Ack...', end='', flush=True)
        while curr_time < time_end and not all(ack.values()):
            curr_time = datetime.datetime.now()
            if curr_time > (time_gw_stops - datetime.timedelta(seconds=1)):  # 1 second buffer
                new_bridges = self.get_relevant_bridges(gateway_id)
                if bridges != new_bridges:
                    bridges = new_bridges
                    debug_print(f'Bridges seen by GW changed while broadcasting! now broadcasting to {bridges}')
                self.broadcast_calib_mgmt_packet(payload, gateway_id, silent=True, seconds_to_send=SEC_TO_SEND)
                time_gw_stops = datetime.datetime.now() + datetime.timedelta(seconds=SEC_TO_SEND)
                print('!', end='')
                sys.stdout.flush()
            time.sleep(0.5)
            print('.', end='')
            sys.stdout.flush()
            ack = self.check_calib_mgmt_broadcast_received(gateway_id, bridges, payload, time_sent)
        print()  # enter newline
        if not all(ack.values()):
            # list of bridges which did not return ack
            brgs_missed = list(dict(filter(lambda x: x[1] is False, ack.items())).keys())
            debug_print(f'Packet not received by bridges {brgs_missed} after {minutes_timeout} minutes!')
        else:
            debug_print(
                f'Ack Received after {datetime.datetime.now() - time_sent} by bridges {list(ack.keys())}')
        return ack


from math import exp
import sys
from wiliot_deployment_tools.automatic_configuration_tool.automatic_configuration_tool import AutomaticConfigurationTool
from argparse import ArgumentParser
from wiliot_core import check_user_config_is_ok


def main():
    parser = ArgumentParser(prog='wlt-config', 
                            description='Automatic Configuration Tool - CLI Tool to automatically configure Wiliot deployments',
                            epilog='example usage: wlt-config -owner wiliot -location "My Deployment" -ota -pacing_interval 10 -ignore_bridges 1234ABCD0123')
    
    # Required Arguments
    required = parser.add_argument_group('required arguments')
    required.add_argument("-owner", type=str, help="Platform owner id", required=True)
    required.add_argument("-location", type=str, help="Location name in Wiliot platform. If location has ' ' in name, input location wrapped with double quotes: -location \"LOCATION NAME\" ", required=True)
    
    # Optional Arguments
    optional = parser.add_argument_group('additional (optional) arguments')
    optional.add_argument("-ota", action='store_true', help="Updating FW version to latest for all devices")
    optional.add_argument("-no_gp_zone", action='store_false',
                        help="Don't use global pacing group by zone (all bridges will be GPG=0)")
    optional.add_argument("-pacer", type=int, help="Pacing interval for all devices", default=15)
    optional.add_argument("-ignore_bridges", type=str, help="Bridges to ignore in the tool - their configuration won't "
                                                           "be changed", nargs="*", default=[])
    optional.add_argument('-expected_num_brgs', type=int, help='Number of expected bridges in location. The tool will try to connect to all bridges (excluding those specified in ignore bridges) until reaching expected number.', required=False)

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)
    args = parser.parse_args()
    user_config_file_path, asset_api_key, is_success = check_user_config_is_ok(args.owner, 'prod', 'asset')
    if is_success:
        print('Asset Management credentials saved/upload from {}'.format(user_config_file_path))
    else:
        raise Exception('invalid Asset Management credentials - please try again to login')
    user_config_file_path, edge_api_key, is_success = check_user_config_is_ok(args.owner, 'prod', 'edge')
    if is_success:
        print('Edge Management credentials saved/upload from {}'.format(user_config_file_path))
    else:
        raise Exception('invalid Edge Management credentials - please try again to login')

    act = AutomaticConfigurationTool(edge_api_key=edge_api_key, asset_api_key=asset_api_key, owner=args.owner,
                                     location=args.location, env='prod', ota_upgrade=args.ota,
                                     ignore_bridges=args.ignore_bridges, pacing_interval=args.pacer,
                                     use_gp_zone=args.no_gp_zone, expected_num_brgs=args.expected_num_brgs)
    act.init_stage()

def main_cli():
    main()

if __name__ == '__main__':
    main()




from argparse import ArgumentParser
import time
from wiliot_deployment_tools.interface.ble_simulator import BLESimulator

def main():
    parser = ArgumentParser(prog='wlt-ble-sim',
                            description='BLE Simulator - CLI Tool to use GW BLE Simulator')
    required = parser.add_argument_group('required arguments')
    required.add_argument('-p', type=str, help="port", required=True)
    required.add_argument('-c', type=str, help="channel", required=True)
    required.add_argument('-dup', type=str, help="duplicates", required=True)
    required.add_argument('-delay', type=str, help="delay", required=True)
    required.add_argument('-output', type=str, help="output power", required=True)
    required.add_argument('-packet', type=str, help="packet", required=True)
    required.add_argument('-ts', type=float, help="time stamp") 

    
    args = parser.parse_args()
    b = BLESimulator(args.p)
    b.set_sim_mode(True)
    b.trigger_by_time_stamp(args.ts)
    b.send_packet(args.packet, args.dup, args.output, args.c, args.delay)
    b.set_sim_mode(False)
    
def main_cli():
    main()


if __name__ == '__main__':
    main()



import datetime
import json
import os
import time
import pandas as pd
import numpy as np
import plotly.express as px
import jinja2

from wiliot_api.api_client import WiliotCloudError
from wiliot_deployment_tools.api.extended_api import ExtendedEdgeClient
from wiliot_deployment_tools.common.debug import debug_print
from wiliot_deployment_tools.interface.ble_sniffer import BLESniffer, BLESnifferContext
from wiliot_deployment_tools.interface.if_defines import RX_CHANNELS
from wiliot_deployment_tools.gw_certificate.tests.static.downlink_defines import *
from wiliot_deployment_tools.interface.mqtt import MqttClient
from wiliot_deployment_tools.interface.pkt_generator import BrgPktGenerator
from wiliot_deployment_tools.gw_certificate.tests.generic import PASS_STATUS, GenericTest, GenericStage

class GenericDownlinkStage(GenericStage):
    def __init__(self, sniffer:BLESniffer, mqttc:MqttClient, pkt_gen:BrgPktGenerator, stage_name, **kwargs):
        self.__dict__.update(kwargs)
        super().__init__(stage_name=stage_name, **self.__dict__)        
        
        #Clients
        self.sniffer = sniffer
        self.mqttc = mqttc
        self.pkt_gen = pkt_gen
        
        #Stage Params
        self.sent_pkts = []
        self.sniffed_pkts = pd.DataFrame()
        
        #Paths
        self.sent_csv_path = os.path.join(self.test_dir, f'{self.stage_name}_sent.csv')
        self.sniffed_csv_path = os.path.join(self.test_dir, f'{self.stage_name}_sniffed.csv')
        self.graph_html_path = os.path.join(self.test_dir, f'{self.stage_name}.html')

        
    def prepare_stage(self):
        super().prepare_stage()
        self.sniffer.flush_pkts()
    
    def send_adv_payloads(self, tx_max_durations=TX_MAX_DURATIONS, retries=RETRIES):
        sent_pkts = []
        for tx_max_duration in tx_max_durations:
            debug_print(f'Tx Max Duration {tx_max_duration}')
            for retry in retries:
                self.pkt_gen.increment_all()
                brg_hb = self.pkt_gen.get_brg_hb()
                tx_max_retries = tx_max_duration//100
                sent_payload = self.mqttc.advertise_packet(raw_packet=brg_hb, tx_max_duration=tx_max_duration, use_retries=self.use_retries)
                debug_print(f'{sent_payload} sent to GW')

                sent_payload = json.dumps(sent_payload)
                sent_pkts.append({'tx_max_duration': tx_max_duration, 'tx_max_retries': tx_max_retries,
                                    'retry': retry, 'pkt': brg_hb[12:], 'payload': sent_payload, 'time_sent': datetime.datetime.now()})
                time.sleep(max(MAX_RX_TX_PERIOD_SECS, (tx_max_duration/1000)*1.2))
        time.sleep(10)
        return sent_pkts
        
    def process_sniffed_pkts(self, sent_pkts, sniffer:BLESniffer):
        for pkt in sent_pkts:
            # Get vars from dict
            raw_packet = pkt['pkt']
            # Get packets from sniffer
            sniffed_pkts = sniffer.get_filtered_packets(raw_packet=raw_packet)
            pkt['channel'] = str(sniffer.rx_channel)
            pkt['num_pkts_received'] = len(sniffed_pkts)
            self.sniffed_pkts = pd.concat([self.sniffed_pkts, sniffed_pkts.to_pandas()])
        self.sent_pkts += sent_pkts

    def generate_stage_report(self):
        # Create graph and trendline
        self.sent_pkts = pd.DataFrame(self.sent_pkts)
        x_value = ('tx_max_duration', 'TX Max Duration') if not self.use_retries else ('tx_max_retries', 'TX Max Retries')
        fig = px.scatter(self.sent_pkts, x=x_value[0], y='num_pkts_received', color='channel', title=f'Packets Received by Sniffer / {x_value[1]}',
                         trendline='ols', labels={x_value[0]: x_value[1], 'num_pkts_received': 'Number of packets received', 'channel': "BLE Adv. Channel"})
        fig.update_layout(scattermode="group", scattergap=0.95)
        trendline_info = px.get_trendline_results(fig)
        # Calculate whether stage pass/failed
        for channel, channel_df in trendline_info.groupby('BLE Adv. Channel'):
            channel_pass = False
            channel_pkts = self.sent_pkts[self.sent_pkts['channel'] == channel]
            channel_trendline = channel_df['px_fit_results'].iloc[0]
            slope = channel_trendline.params[1]
            rsquared = channel_trendline.rsquared
            # Determine Channel Pass
            if self.use_retries:
                if (rsquared > 0.81) and (0.9 < slope < 1.1):
                    channel_pass = False
            else:
                if (rsquared > 0.8) and (slope > 0):
                    channel_pass = False
            if channel_pass == False:
                self.stage_pass = False
            self.add_to_stage_report(f"Channel {channel}: {PASS_STATUS[channel_pass]}")
            self.add_to_stage_report(f"- Total {len(channel_pkts['payload'])} MQTT payloads sent")
            self.add_to_stage_report(f"- Total {sum(channel_pkts['num_pkts_received'])} BLE Packets received by sniffer")
            self.add_to_stage_report(f"- R Value: {rsquared} | Slope: {slope}")
        # Export all stage data
        self.sent_pkts.to_csv(self.sent_csv_path)
        self.add_to_stage_report(f'Sent data saved - {self.sent_csv_path}')
        self.sniffed_pkts.to_csv(self.sniffed_csv_path)
        self.add_to_stage_report(f'Sniffed data saved - {self.sniffed_csv_path}')
        fig.write_html(self.graph_html_path)
        self.add_to_stage_report(f'Graph saved - {self.graph_html_path}')
        debug_print(self.report)
        
        # Generate HTML
        graph_div = fig.to_html(full_html=False, include_plotlyjs='cdn')
        self.report_html = self.template_engine.render_template('stage.html', stage_name = self.stage_name, stage_pass=self.stage_pass, stage_report=self.report.split('\n'), graph = graph_div)
        return self.report
        
class InitStage(GenericDownlinkStage):
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)
        super().__init__(**self.__dict__, stage_name=type(self).__name__)

    def run(self):                
        self.start_time = datetime.datetime.now()
        self.stage_pass = True
        for channel in RX_CHANNELS:
            debug_print(f'RX Channel {channel}')
            with BLESnifferContext(self.sniffer, channel) as sniffer:
                # Send the packets
                sent_pkts = self.send_adv_payloads(STAGE_CONFIGS[INIT_STAGE][0], STAGE_CONFIGS[INIT_STAGE][1])
                # Process sniffed packets
                self.process_sniffed_pkts(sent_pkts, sniffer)
        return True
class ExtendedStage(GenericDownlinkStage):
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)
        super().__init__(**self.__dict__, stage_name=type(self).__name__)

    def run(self):                
        self.start_time = datetime.datetime.now()
        self.stage_pass = True
        for channel in RX_CHANNELS:
            debug_print(f'RX Channel {channel}')
            with BLESnifferContext(self.sniffer, channel) as sniffer:
                # Send the packets
                sent_pkts = self.send_adv_payloads(STAGE_CONFIGS[EXTENDED_STAGE][0], STAGE_CONFIGS[EXTENDED_STAGE][1])
                # Process sniffed packets
                self.process_sniffed_pkts(sent_pkts, sniffer)
        return True
    
STAGES = [InitStage, ExtendedStage]

class DownlinkTest(GenericTest):
    def __init__(self, **kwargs):        
        self.__dict__.update(kwargs)
        super().__init__(**self.__dict__, test_name=type(self).__name__)
        self.pkt_gen = BrgPktGenerator()
        self.pkt_gen.set_bridge_id(DEFAULT_BRG_ID)
        self.use_retries = self.topic_suffix == '-test'
        self.stages = [stage(**self.__dict__) for stage in STAGES]
    
    def run(self):
        super().run()
        self.test_pass = True
        for stage in self.stages:
            stage.prepare_stage()
            stage.run()
            self.add_to_test_report(stage.generate_stage_report())
            if stage.stage_pass == False:
                self.test_pass = False



from wiliot_deployment_tools.gw_certificate.tests.coupling import CouplingTest
from wiliot_deployment_tools.gw_certificate.tests.downlink import DownlinkTest 
from wiliot_deployment_tools.gw_certificate.tests.connection import ConnectionTest
from wiliot_deployment_tools.gw_certificate.tests.uplink import UplinkTest
TESTS = [ConnectionTest, UplinkTest, DownlinkTest]



import unittest
from wiliot_deployment_tools.api.extended_api import ExtendedPlatformClient
from api_secrets import *
class TestExtendedPlatform(unittest.TestCase):

    def test_init(self):
        plat = ExtendedPlatformClient(WH_ASSET, WH_OWNERID)
        self.assertIsInstance(plat, ExtendedPlatformClient)
    
    def test_get_locations(self):
        plat = ExtendedPlatformClient(WH_ASSET, WH_OWNERID)
        self.assertIsInstance(plat.get_locations(), list)
    
    def test_get_locations_bridges(self):
        plat = ExtendedPlatformClient(WH_ASSET, WH_OWNERID)
        self.assertIsInstance(plat.get_locations_bridges(), list)

    def test_get_location_id(self):
        plat = ExtendedPlatformClient(WH_ASSET, WH_OWNERID)
        for location in plat.get_locations():
            name = location['name']
            self.assertIsInstance(plat.get_location_id(name), str)
    
    def test_get_location_bridges(self):
        plat = ExtendedPlatformClient(WH_ASSET, WH_OWNERID)
        for location in plat.get_locations():
            name = location['name']
            self.assertIsInstance(plat.get_location_bridges(name), list)

if __name__ == '__main__':
    unittest.main()



import datetime
import time
from wiliot_deployment_tools.common.debug import debug_print
from wiliot_deployment_tools.gw_certificate.api_if.gw_capabilities import GWCapabilities
from wiliot_deployment_tools.gw_certificate.tests.generic import PASS_STATUS, GenericStage, GenericTest
from wiliot_deployment_tools.gw_certificate.api_if.api_validation import validate_message, MESSAGE_TYPES
from wiliot_deployment_tools.interface.mqtt import MqttClient

class ConnectionStage(GenericStage):
    def __init__(self, mqttc:MqttClient, **kwargs):
        self.mqttc = mqttc
        self.__dict__.update(kwargs)
        super().__init__(stage_name=type(self).__name__, **self.__dict__)
        
    def run(self):
        self.stage_pass = False
        input('Please unplug GW from power. Press enter when unplugged')
        self.mqttc.flush_messages()
        input('Please plug GW back to power. Press enter when plugged')
        debug_print('Waiting for GW to connect... (Timeout 3 minutes)')
        timeout = datetime.datetime.now() + datetime.timedelta(minutes=3)
        while datetime.datetime.now() < timeout and not self.mqttc.check_gw_seen():
            pass
        if self.mqttc.check_gw_seen():
            status_pkts = self.mqttc.get_all_messages_from_topic('status')
            if len(status_pkts) > 0:
                status_pkt = status_pkts[0].body_ex
                self.add_to_stage_report('GW Status packet received:')
                debug_print(status_pkt)
                validation = validate_message(MESSAGE_TYPES.STATUS, status_pkt)
                self.stage_pass = validation[0]
                # set GW Capabilities:
                for key, value in status_pkt.items():
                    if key in GWCapabilities.get_capabilities() and type(value) is bool:
                        self.gw_capabilities.set_capability(key, value)
                        self.add_to_stage_report(f'Set Capability: {key} - {value}')
                # Add reason test failed to report if neccessary
                self.add_to_stage_report(validation[1]) if not self.stage_pass else None
                # Add status packet to test report
                self.add_to_stage_report(str(status_pkt))
        else:
            self.add_to_stage_report("No message recieved from GW in status topic after 3 mins")
    
    def generate_stage_report(self):
        self.report_html = self.template_engine.render_template('stage.html', stage_name = self.stage_name, stage_pass=self.stage_pass, stage_report=self.report.split('\n'))
        debug_print(self.report)
        return super().generate_stage_report()
    

class ConnectionTest(GenericTest):
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)
        super().__init__(**self.__dict__, test_name=type(self).__name__)
        self.stages = [ConnectionStage(**self.__dict__)]
        
    def run(self):
        super().run()
        self.test_pass = False
        for stage in self.stages:
            stage.prepare_stage()
            stage.run()
            self.test_pass = stage.stage_pass
            self.add_to_test_report(stage.generate_stage_report())



from dataclasses import dataclass
from typing import Literal

@dataclass
class GWCapabilities:
    tagMetadataCouplingSupported: bool = False
    downlinkSupported: bool = False
    bridgeOtaUpgradeSupported: bool = False
    fwUpgradeSupported: bool = False
    geoLocationSupport: bool = False
    
    @staticmethod
    def get_capabilities():
        return list(GWCapabilities.__dataclass_fields__.keys())
    
    def get_capability(self, capability):
        assert capability in GWCapabilities.get_capabilities(), f'{capability} is not a valid capability'
        return getattr(self, capability)
    
    def set_capability(self, capability, value:bool):
        assert capability in GWCapabilities.get_capabilities(), f'{capability} is not a valid capability'
        setattr(self, capability, value)


import copy
import datetime
from enum import Enum
import json
import logging
import time
from typing import Literal
import paho.mqtt.client as mqtt
import ssl
from wiliot_deployment_tools.ag.ut_defines import PACKETS, PAYLOAD, MGMT_PKT, SIDE_INFO_PKT, GW_ID, NFPKT
from wiliot_deployment_tools.ag.wlt_types_ag import GROUP_ID_BRG2GW, GROUP_ID_SIDE_INFO
from wiliot_deployment_tools.ag.wlt_types_data import DATA_DEFAULT_GROUP_ID, DataPacket
from wiliot_deployment_tools.common.debug import debug_print

DATA_PKT = 'data_pkt'

class GwAction(Enum):
    DISABLE_DEV_MODE = "DevModeDisable"
    REBOOT_GW ="rebootGw"
    GET_GW_INFO ="getGwInfo"

class WltMqttMessage:
    def __init__(self, body, topic):
        self.body = body
        self.mqtt_topic = topic
        self.mqtt_timestamp = datetime.datetime.now()
        self.body_ex = copy.deepcopy(body)
        if "data" in self.mqtt_topic:
            for pkt in self.body_ex[PACKETS]:
                data_pkt = DataPacket()
                data_pkt.set(pkt[PAYLOAD])
                if data_pkt.pkt != None:
                    if data_pkt.hdr.group_id == GROUP_ID_BRG2GW:
                        pkt[MGMT_PKT] = copy.deepcopy(data_pkt)
                if data_pkt.hdr.group_id == GROUP_ID_SIDE_INFO:
                    pkt[SIDE_INFO_PKT] = copy.deepcopy(data_pkt)
                if data_pkt.hdr.group_id == DATA_DEFAULT_GROUP_ID:
                    pkt[DATA_PKT] = copy.deepcopy(data_pkt)
                    pass

    def __repr__(self) -> str:
        if self.body_ex != {}:
            return str(self.body_ex)
        return str(self.body)


class WltMqttMessages:
    def __init__(self):
        self.data = []
        self.status = []
        self.update = []
        self.all = []

    def insert(self, pkt):
        self.all.append(pkt)
        if "data" in pkt.mqtt_topic:
            self.data.append(pkt)
        elif "status" in pkt.mqtt_topic:
            self.status.append(pkt)
        elif "update" in pkt.mqtt_topic:
            self.update.append(pkt)
            
    def __repr__(self) -> str:
        return f'Data {self.data} \n Status {self.status} \n Update {self.update}'
            
class MqttClient:

    def __init__(self, gw_id, owner_id, logger_filepath=None, topic_suffix=''):
        # Set variables
        self.gw_id = gw_id
        self.owner_id = owner_id
        
        
        # Configure logger
        logger = logging.getLogger('mqtt')
        logger.setLevel(logging.DEBUG)
        if logger_filepath is not None:
            # create file handler which logs even debug messages
            fh = logging.FileHandler(logger_filepath)
            fh.setLevel(logging.DEBUG)
            logger.addHandler(fh)
            debug_print(f'MQTT Logger initialized at {logger_filepath}')
        self.logger = logger
        
        # Configure Paho MQTT Client
        client_id = '{}-republish'.format('gw_id')
        self.userdata = {'messages': WltMqttMessages(), 'gw_seen': False , 'logger': self.logger
            }
        self.client = mqtt.Client(client_id, userdata=self.userdata)
        self.client.on_message = on_message
        self.client.on_connect = on_connect
        self.client.on_disconnect = on_disconnect
        self.client.on_subscribe = on_subscribe
        self.client.on_unsubscribe = on_unsubscribe
        self.client.on_log = on_log
        self.client.tls_set(tls_version=ssl.PROTOCOL_TLSv1_2)
        debug_print(f'Connecting to MQTT broker: tls://broker.hivemq.com:8883, Keepalive=60')
        self.client.connect("broker.hivemq.com", port=8883, keepalive=60)
        # Set Topics
        self.update_topic = f"update{topic_suffix}/{owner_id}/{gw_id}"
        debug_print(f'Subscribe to {self.update_topic}...')
        self.client.subscribe(self.update_topic)
        self.data_topic = f"data{topic_suffix}/{owner_id}/{gw_id}"
        debug_print(f'Subscribe to {self.data_topic}...')
        self.client.subscribe(self.data_topic)
        self.status_topic = f"status{topic_suffix}/{owner_id}/{gw_id}"
        debug_print(f'Subscribe to {self.status_topic}...')
        self.client.subscribe(self.status_topic)
        self.client.loop_start()
        while(not self.client.is_connected()):
            debug_print(f'Waiting for MQTT connection...')
            time.sleep(1)
        debug_print('Connected to MQTT.')
    
    # Downstream Interface
    def send_action(self, action:GwAction):
        """
        Send an action to the gateway
        :param action: GwAction - Required
        """
        assert isinstance(action, GwAction), 'Action Must be a GWAction!'
        message_info = self.client.publish(self.update_topic, payload=json.dumps({"action": action.value}))
        message_info.wait_for_publish()
        return message_info
    
    def send_payload(self, payload, topic:Literal['update', 'data', 'status']='update'):
        """
        Send a payload to the gateway
        :type payload: dict
        :param payload: payload to send
        :type topic: Literal['update', 'data', 'status']
        :param topic: defualts to update
        """
        topic = {'update': self.update_topic,
                 'data': self.data_topic, 
                 'status': self.status_topic}[topic]
        message_info = self.client.publish(topic, payload=json.dumps(payload))
        message_info.wait_for_publish()
        return message_info
    
    def advertise_packet(self, raw_packet, tx_max_duration=800, use_retries=False):
        if len(raw_packet) < 62:
            if len(raw_packet) == 54:
                raw_packet = 'C6FC' + raw_packet
            if len(raw_packet) == 58:
                raw_packet = '1E16' + raw_packet
        if len(raw_packet) > 62:
            raw_packet = raw_packet[-62:]
        
        assert len(raw_packet) == 62, 'Raw Packet must be 62 chars long!'

        if use_retries:
            payload = {
                'action': 0, # Advertise BLE Packet
                'txPacket': raw_packet, # Raw Packet
                'txMaxRetries': tx_max_duration / 100, # Tx Max Retries
            }
        else:
            payload = {
            'txPacket': raw_packet, # Raw Packet
            'txMaxDurationMs': tx_max_duration, # Tx Max Duration
            'action': 0 # Advertise BLE Packet
        }
        
        self.send_payload(payload, topic='update')
        return payload

    def check_gw_seen(self):
        return self.userdata['gw_seen']
    
    def get_gw_info(self):
        self.flush_messages()
        self.send_action(GwAction.GET_GW_INFO)
        time.sleep(5)
        debug_print('---GW INFO---')
        try:
            gw_info = self.userdata['messages'].status[0]
            debug_print(gw_info)
            self.flush_messages()
            return gw_info
        except IndexError:
            debug_print('No GW INFO')
            self.flush_messages()
            return False
        
    def get_gw_configuration(self):
        self.flush_messages()
        self.send_action(GwAction.REBOOT_GW)
        debug_print('---GW CONFIG---')
        try:
            debug_print(self.userdata['messages'].status)
            return True
        except KeyError:
            return False
            
    def exit_custom_mqtt(self, mqtt_mode:Literal['automatic', 'manual' ,'legacy']):
        if mqtt_mode == 'legacy':
            return self.send_action(GwAction.DISABLE_DEV_MODE)
        elif mqtt_mode == 'automatic':
            custom_mqtt = {
                "custom-mqtt": False,
                "broker_url": "broker.hivemq.com",
                "port": 8883,
                "username": "",
                "password": "",
                "update-topic": f"update/{self.owner_id}/{self.gw_id}",
                "status-topic": f"status/{self.owner_id}/{self.gw_id}",
                "data-topic": f"data/{self.owner_id}/{self.gw_id}"
                }
            return self.send_payload(custom_mqtt)
        elif mqtt_mode == 'manual':
            debug_print(f"Make sure GW {self.gw_id} is set to Wiliot MQTT broker")
            return True
    
    # Packet Handling
    def flush_messages(self):
        self.userdata = {'messages': WltMqttMessages(), 'gw_seen': False , 'logger': self.logger}
        self.client.user_data_set(self.userdata)
    
    def get_all_messages_from_topic(self, topic:Literal['status', 'data', 'update']):
        return getattr(self.userdata['messages'], topic)
    
    def get_all_pkts_from_topic(self, topic:Literal['status', 'data', 'update']):
        pkts = []
        if topic == 'data':
            for p in eval(f'self.userdata["messages"].{topic}'):
                gw_id = p.body_ex[GW_ID] if GW_ID in p.body_ex else ""
                if PACKETS in p.body_ex:
                    for pkt in p.body_ex[PACKETS]:
                        pkt[GW_ID] = gw_id
                        pkts += [pkt]
            return pkts
    
    def get_coupled_tags_pkts(self):
        return [p for p in self.get_all_pkts_from_topic('data') if NFPKT in p]
    
    def get_uncoupled_tags_pkts(self):
        return [p for p in self.get_all_pkts_from_topic('data') if NFPKT not in p]
    
    def get_all_tags_pkts(self):
        return [p for p in self.get_all_pkts_from_topic('data')]




# MQTT Client callbacks

def on_connect(mqttc, userdata, flags, rc):
    rc_dict = {0: 'MQTT: Connection successful',
               1: 'MQTT: Connection refused - incorrect protocol version',
               2: 'MQTT: Connection refused - invalid client identifier',
               3: 'MQTT: Connection refused - server unavailable',
               4: 'MQTT: Connection refused - bad username or password',
               5: 'MQTT: Connection refused - not authorised'}
    try:
        message = rc_dict[rc]
    except KeyError:
        message = f'MQTT: Cannot connect, RC:{rc}'
    userdata['logger'].info(message)


def on_disconnect(mqttc, userdata, rc):
    if rc != 0:
        userdata['logger'].info(f"MQTT: Unexpected disconnection. {rc}")
    else:
        userdata['logger'].info('MQTT: Disconnect')


def on_subscribe(mqttc, userdata, mid, granted_qos):
    userdata['logger'].info(f"MQTT: Subscribe, MessageID {mid}, QoS {granted_qos}")


def on_unsubscribe(mqttc, userdata, mid):
    userdata['logger'].info(f"MQTT: Unsubscribe, MessageID {mid}")

def on_message(mqttc, userdata, message):
    data = json.loads(message.payload.decode("utf-8"))
    userdata['messages'].insert(WltMqttMessage(data, message.topic))
    userdata['logger'].debug(f'{message.topic}: {data}')
    if(userdata['gw_seen'] is False):
        userdata['gw_seen'] = True
    
def on_log(client, userdata, level, buf):
    # MQTT_LOG_INFO = 0x01
    # MQTT_LOG_NOTICE = 0x02
    # MQTT_LOG_WARNING = 0x04
    # MQTT_LOG_ERR = 0x08
    # MQTT_LOG_DEBUG = 0x10
    if (level < mqtt.MQTT_LOG_DEBUG):
        userdata['logger'].info(f"mqtt_log: level={level}, msg={buf}")


import os
import json
from jsonschema import validate
from wiliot_deployment_tools.ag.ut_defines import GW_CONF, GW_API_VERSION
import pkg_resources

from enum import Enum

class MESSAGE_TYPES(Enum):
    STATUS = "status"
    DATA = "data"
    LOGS = "logs"
FALLBACK_API = "200"

def validate_message(message_type: MESSAGE_TYPES, message: dict) -> tuple[bool, str]:
    """
    Validate MQTT message
    :type message_type: MESSAGE_TYPES
    :param message_type: MQTT message type
    :type message: dict
    :param message: MQTT Message
    :return: tuple (bool, str)
    """
    try:
        api_version = message[GW_CONF][GW_API_VERSION]
    except KeyError:
        api_version = FALLBACK_API
    json_path = pkg_resources.resource_filename(__name__, f"{api_version}/{message_type.value}.json")
    with open(json_path) as f:
        schema = json.load(f)
    try:
        validate(message, schema)
        return (True, None)
    except Exception as e:
        return (False, e.message)


from typing import Literal
from wiliot_deployment_tools.ag.wlt_types import WltPkt
from wiliot_deployment_tools.ag.ut_defines import *
from wiliot_deployment_tools.ag.wlt_types_ag import *

DATA_DEFAULT_GROUP_ID = 0x020000

class GenericPacket():
    def __init__(self, payload=0, pkt_id=0):
        self.payload = payload
        self.pkt_id = pkt_id

    def __eq__(self, other):
        if isinstance(other, GenericV7):
            return (
                self.payload == other.payload and
                self.pkt_id == other.pkt_id
            )
        return False

    def dump(self):
        string = bitstruct.pack("u160u32", self.payload, self.pkt_id)
        return string.hex().upper()

    def set(self, string):
        d = bitstruct.unpack("u160u32", binascii.unhexlify(string))
        self.payload = d[0]
        self.pkt_id = d[1]


class DataPacket(WltPkt):
    def __init__(self, group_id=None, service_uuid:Literal["tag", "bridge"]='bridge', *args, **kwargs):
        if service_uuid == 'bridge':
            uuid_msb=HDR_DEFAULT_BRG_UUID_MSB
            uuid_lsb=HDR_DEFAULT_BRG_UUID_LSB
        else:
            uuid_msb=HDR_DEFAULT_TAG_UUID_MSB
            uuid_lsb=HDR_DEFAULT_TAG_UUID_LSB
        super().__init__(hdr=Hdr(uuid_msb=uuid_msb, uuid_lsb=uuid_lsb), generic=GenericPacket(), *args, **kwargs)
        if group_id is None:
            self.hdr.group_id = DATA_DEFAULT_GROUP_ID
        else:
            self.hdr.group_id = group_id
    
    def set(self, string):
        if not string.startswith("1E16"):
            string = "1E16" + string
        self.hdr.set(string[0:14])
        if self.hdr.group_id == DATA_DEFAULT_GROUP_ID:
            self.generic = GenericPacket()
            self.generic.set(string[14:62])
        else:
            super().set(string)
            
    def dump(self):
        return super().dump()
    
    def __repr__(self) -> str:
        return self.dump()


# External Imports
import datetime
import os
from typing import Literal
import webbrowser
from jinja2 import Environment, FileSystemLoader
import pkg_resources

# Internal Imports
from wiliot_deployment_tools.common.analysis_data_bricks import initialize_logger
from wiliot_deployment_tools.common.debug import debug_print
from wiliot_core.utils.utils import WiliotDir
from wiliot_deployment_tools.interface.ble_sniffer import BLESniffer
from wiliot_deployment_tools.interface.uart_if import UARTError, UARTInterface
from wiliot_deployment_tools.interface.ble_simulator import BLESimulator
from wiliot_deployment_tools.interface.mqtt import MqttClient
from wiliot_deployment_tools.gw_certificate.tests import *
from wiliot_deployment_tools.interface.uart_ports import get_uart_ports
from wiliot_deployment_tools.gw_certificate.api_if.gw_capabilities import GWCapabilities
class GWCertificateError(Exception):
    pass


class TemplateEngine:
    def __init__(self):
        self.template_dir = pkg_resources.resource_filename(__name__, 'templates')
        self.env = Environment(loader=FileSystemLoader(self.template_dir))

    def get_template(self, template_name):
        return self.env.get_template(template_name)

    def render_template(self, template_name, **kwargs):
        template = self.get_template(template_name)
        return template.render(**kwargs)


class GWCertificate:
    def __init__(self, gw_id, owner_id, tests:list = [], coupled=False, topic_suffix=''):
        # Runtime
        self.env_dirs = WiliotDir()
        self.current_datetime = datetime.datetime.now().strftime('%Y%m%d_%H%M%S')
        self.certificate_dir = os.path.join(self.env_dirs.get_wiliot_root_app_dir(), 'gw-certificate', self.current_datetime)
        self.env_dirs.create_dir(self.certificate_dir)
        self.start_timestamp = initialize_logger(self.certificate_dir)
        self.logger_filepath = os.path.join(self.certificate_dir, f'{self.start_timestamp}.log')
        self.mqtt_logger_filepath = os.path.join(self.certificate_dir, f'{self.start_timestamp}_mqtt.log')
        self.result_html_path = os.path.join(self.certificate_dir, f'results_{self.current_datetime}.html')
        self.template_engine = TemplateEngine()
        
        # GW-related
        self.gw_id = gw_id
        self.owner_id = owner_id
        self.coupled = coupled
        self.topic_suffix = topic_suffix
        self.mqttc = MqttClient(gw_id, owner_id, self.mqtt_logger_filepath, topic_suffix=topic_suffix)
        self.gw_capabilities = GWCapabilities()
        
        # UART-related
        self.uart_comports = get_uart_ports()
        debug_print(f'UART Ports:{self.uart_comports}')
        if len(self.uart_comports) < 1:
            raise GWCertificateError('A Wiliot GW needs to be connected to USB!')
        self.uart = None
        for port in self.uart_comports:
            try:
                self.uart = UARTInterface(port)
                break
            except UARTError as e:
                debug_print(f'Port: {port} - {e}')
        if type(self.uart) is not UARTInterface:
            raise GWCertificateError("Cannot initialize any port!")
        self.ble_sim = BLESimulator(self.uart)
        self.sniffer = BLESniffer(self.uart)
        
        tests_to_init = TESTS if tests == [] else tests
        # Tests
        self.tests = [t(**self.__dict__) for t in tests_to_init]
        debug_print(f'Running Tests: {self.tests}')
    
    def run_tests(self):
        finished_tests = []
        for test in self.tests:
            # # Run only needed Coupling/Uplink test
            # if (type(test) == CouplingTest and self.gw_capabilities.get_capability('tagMetadataCouplingSupported') is False) or \
            #     type(test) == UplinkTest and self.gw_capabilities.get_capability('tagMetadataCouplingSupported') is True:
            #         continue
            test.run()
            finished_tests.append(test)
        for test in finished_tests:
            test.end_test()
        
    def create_results_html(self):
        with open(self.logger_filepath, 'r') as f:
            log = f.read().split('\n')
        with open(self.mqtt_logger_filepath, 'r') as f:
            mqtt_log = f.read().split('\n')
        
        html = self.template_engine.render_template('results.html', tests=self.tests,
                                                    log = log,
                                                    mqtt_log = mqtt_log,
                                                    gw_id = self.gw_id,
                                                    datetime = self.current_datetime)
        with open(self.result_html_path, 'w') as f:
            f.write(html)
        debug_print("Test Finished. Results HTML Saved: " + self.result_html_path)
        webbrowser.open('file://' + os.path.realpath(self.result_html_path))

if __name__ == "__main__":
    from api_secrets import *
    gwc = GWCertificate(gw_id='ac233ffb264b',owner_id=832742983939, topic_suffix='-prod', coupled=False,
                        tests=[ConnectionTest,])
    gwc.run_tests()
    gwc.create_results_html()

#TODO - Add manual mode to specify gw capabilities



from wiliot_deployment_tools.api.extended_api import ExtendedEdgeClient, GatewayAction
from wiliot_deployment_tools.common.debug import debug_print

class GatewaySsidControl(ExtendedEdgeClient):
    def toggle_gw_ssid(self, gateway_id):
        """
        Toggles gateway preferred SSID (if a new SSID was not added, function will have no effect)
        :type gateway_id: str
        :param gateway_id: Gateway ID
        :return: True if the cloud successfully sent the action to the gateway, False otherwise
        """
        res = self.send_action_to_gateway(gateway_id, GatewayAction.TOGGLE_SSID)
        debug_print(f'Rebooting GW {gateway_id}')
        return res

    def add_ssid_to_gw(self, gateway_id, ssid, pswd=None):
        """
        Adds SSID/PSWD pair to gateway
        :type gateway_id: str
        :param gateway_id: Gateway ID
        :type ssid: str
        :param ssid: SSID
        :type pswd: str
        :param pswd: Password
        :rtype: bool
        :return: True if the cloud successfully sent the action to the gateway, False otherwise
        """
        return self.send_action_to_gateway(gateway_id, GatewayAction.ADD_SSID, ssid=ssid, pswd=pswd)



import json
import boto3
from wiliot_deployment_tools.common.analysis_data_bricks import get_secret
from wiliot_deployment_tools.common.debug import is_databricks
bucket_name = 'wiliot-gateway-versions-non-prod'
class S3Client():
    def __init__(self):
        if is_databricks():
            access_key = get_secret('warehouse', 'AWS_ACCESS_KEY_ID')
            secret_key = get_secret('warehouse', 'AWS_SECRET_ACCESS_KEY')
            self.s3 = boto3.resource('s3', aws_access_key_id=access_key, aws_secret_access_key=secret_key)
        else:
            self.s3 = boto3.resource('s3')
        self.bucket = self.s3.Bucket(bucket_name)   
    
    def get_gw_validation_json(self, gw_type, version):
        item = self.s3.Object(bucket_name, f'v2/{gw_type}/validation/gw_validation/{version}.json')
        return json.loads(item.get()['Body'].read().decode('utf-8'))['properties']
        
    def get_brg_validation_json(self, gw_type, version):
        item = self.s3.Object(bucket_name, f'v2/{gw_type}/validation/bridge_validation/{version}.json')
        return json.loads(item.get()['Body'].read().decode('utf-8'))['properties']


import logging
from importlib import reload
import os
import shutil
import subprocess
import datetime
from itertools import combinations, chain
from time import sleep
from pathlib import Path
import inspect
from appdirs import user_data_dir

import plotly.express as px
import pandas as pd
import numpy as np
from plotly.offline import plot
from wiliot_deployment_tools.common.debug import debug_print, is_databricks
from os.path import exists
import tabulate
try:
    from wiliot_deployment_tools.internal.utils.network_packet_list import NetworkPacketList, network_data_preparation
    print('NetworkPacketList imported (running internal version)')
except ImportError:
    pass
from wiliot_deployment_tools.common.utils import convert_timestamp_to_datetime, current_timestamp, mstimestamp_to_timezone


if is_databricks():
    logging.getLogger("py4j.java_gateway").setLevel(logging.ERROR)

class WiliotTableError(Exception):
    pass

class WiliotDatabricksUtils:

    def __init__(self, spark):
        self.spark = spark
        self.is_databricks = is_databricks()

    def get_seen_tags(self, table, start_time, end_time, tags_list=None, bridges_list=None, gateways_list=None):
        """
        does an SQL query of the packet data table between specified timestamps
        filters the data for specified tags/bridges/gateways (if specified)
        returns relevant values from the table
        :type table: str
        :param table: name of data table
        :type start_time: float
        :param start_time: time filter start timestamp (UTC milliseconds)
        :type end_time: float
        :param end_time: time filter end timestamp (UTC milliseconds)
        :type tags_list: list
        :param tags_list: list of tags to filter from the data
        :type bridges_list: list
        :param bridges_list: list of bridges to filter from the data
        :type gateways_list: list
        :param gateways_list: list of gateways to filter from the data
        :rtype: pandas DataFrame
        :return: dataframe of data from table
        """
        # TODO - debug the SQL Query with different types of lists
        if bridges_list is not None:
            if len(bridges_list) == 1:
                bridges_list = list(bridges_list) + [""]
            bridges_list = tuple(bridges_list)
        if gateways_list is not None:
            if len(gateways_list) == 1:
                gateways_list = list(gateways_list) + [""]
            gateways_list = tuple(gateways_list)
        if tags_list is not None:
            if len(tags_list) == 1:
                tags_list = list(tags_list) + [""]
            tags_list = tuple(tags_list)
        # adding search by date to improve search time (most tables in data bricks partition is by date)
        query_start_datetime = convert_timestamp_to_datetime(str(start_time))
        query_end_datetime = convert_timestamp_to_datetime(str(end_time))
        start_datetime = query_start_datetime - datetime.timedelta(hours=24)
        end_datetime = query_end_datetime + datetime.timedelta(hours=24)
        start_date = datetime.datetime.strftime(start_datetime, '%Y%m%d')
        end_date = datetime.datetime.strftime(end_datetime, '%Y%m%d')
        sql_method = f"""
        select
            t.* EXCEPT(isLate, json,time)
        from {table} AS t      
        where date between {start_date} and {end_date}
        and timestamp between {start_time} and {end_time}
        """
        if tags_list is not None and tags_list != ():
            sql_method = sql_method + f"""and externalId in {tags_list}
        """
        if bridges_list is not None and bridges_list != ():
            sql_method = sql_method + f"""and bridgeId in {bridges_list}
        """
        if gateways_list is not None and gateways_list != ():
            sql_method = sql_method + f"""and gatewayId in {gateways_list}
        """
        if self.spark is not None:
            debug_print('Running SQL query...', center=True)
            debug_print(sql_method)
            query_data = self.spark.sql(sql_method)
            query_data = query_data.toPandas()
            return query_data
        else:
            raise EnvironmentError("SQL query can only run in databricks")

    def get_seen_events(self, table, start_time, end_time, bridges_list=None, gateways_list=None, platform=False):
        """
        does an SQL query of the packet data table between specified timestamps
        filters the data for specified tags/bridges/gateways (if specified)
        returns relevant values from the table
        :type table: str
        :param table: name of data table
        :type start_time: float
        :param start_time: time filter start timestamp (UTC milliseconds)
        :type end_time: float
        :param end_time: time filter end timestamp (UTC milliseconds)
        :type bridges_list: list
        :param bridges_list: list of bridges to filter the data by
        :type gateways_list: list
        :param gateways_list: list of gateways to filter the data by
        :type platform: bool
        :param platform: platform/management (true/false)
        :rtype: pandas DataFrame
        :return: dataframe of data from table
        """
        # TODO - debug the SQL Query with different types of lists
        devices_list = ()
        if bridges_list is not None:
            if len(bridges_list) == 1:
                devices_list = list(bridges_list) + [""]
            devices_list = tuple(devices_list)
        if gateways_list is not None:
            if len(gateways_list) == 1 and devices_list == ():
                devices_list = list(gateways_list) + [""]
            elif devices_list is not None:
                devices_list = devices_list + tuple(gateways_list)
            devices_list = tuple(devices_list)

        # adding search by date to improve search time (most tables in data bricks partition is by date)
        query_start_datetime = convert_timestamp_to_datetime(str(start_time))
        query_end_datetime = convert_timestamp_to_datetime(str(end_time))
        start_datetime = query_start_datetime - datetime.timedelta(hours=36)
        end_datetime = query_end_datetime + datetime.timedelta(hours=36)
        start_date = datetime.datetime.strftime(start_datetime, '%Y%m%d')
        end_date = datetime.datetime.strftime(end_datetime, '%Y%m%d')
        sql_method = f"""
        select *
        from {table}
        where date between {start_date} and {end_date} 
        and start between {start_time} and {end_time} 
        """
        if not platform:
            connectivity_event = "name = 'NTWK'"
            id_filter = "id"
        else:
            connectivity_event = "eventName = 'connectivity'"
            id_filter = "assetId"

        if devices_list is not None:
            sql_method = sql_method + f"""and ({connectivity_event} or {id_filter} in {devices_list})
        """
        else:
            sql_method = sql_method + f"""and {connectivity_event}"""
        if self.spark is not None:
            debug_print('Running SQL query...', center=True)
            debug_print(sql_method)
            query_data = self.spark.sql(sql_method)
            query_data = query_data.toPandas()
            return query_data
        else:
            raise EnvironmentError("SQL query can only run in databricks")

    def get_sequence_id_data(self, table, start_time, end_time, gateways_list=None):
        """
        does an SQL query of the packet data table between specified timestamps
        returns only sequence id
        :type table: str
        :param table: name of data table
        :type start_time: float
        :param start_time: time filter start timestamp (UTC milliseconds)
        :type end_time: float
        :param end_time: time filter end timestamp (UTC milliseconds)
        :type gateways_list: list
        :param gateways_list: list of gateways to filter the data by
        :rtype: pandas DataFrame
        :return: dataframe of data from table
        """

        if gateways_list is not None:
            if len(gateways_list) == 1:
                gateways_list = list(gateways_list) + [""]
            gateways_list = tuple(gateways_list)
        # adding search by date to improve search time (most tables in data bricks partition is by date)
        query_start_datetime = convert_timestamp_to_datetime(str(start_time))
        query_end_datetime = convert_timestamp_to_datetime(str(end_time))
        start_datetime = query_start_datetime - datetime.timedelta(hours=36)
        end_datetime = query_end_datetime + datetime.timedelta(hours=36)
        start_date = datetime.datetime.strftime(start_datetime, '%Y%m%d')
        end_date = datetime.datetime.strftime(end_datetime, '%Y%m%d')
        sql_method = f"""
        select sequenceId, timestamp, gatewayId, gatewayName
        from {table}
        where date between {start_date} and {end_date} 
        and timestamp between {start_time} and {end_time} 
        """
        if gateways_list is not None:
            sql_method = sql_method + f"""and gatewayId in {gateways_list}
        """
        if self.spark is not None:
            debug_print('Running SQL query...', center=True)
            debug_print(sql_method)
            query_data = self.spark.sql(sql_method)
            query_data = query_data.toPandas()
            return query_data
        else:
            raise EnvironmentError("SQL query can only run in databricks")

    def get_statistics_data(self, table, start_time, end_time, gateways_list=None):
        """
        does an SQL query of the statistics data table between specified timestamps
        :type table: str
        :param table: name of data table
        :type start_time: float
        :param start_time: time filter start timestamp (UTC milliseconds)
        :type end_time: float
        :param end_time: time filter end timestamp (UTC milliseconds)
        :type gateways_list: list
        :param gateways_list: list of gateways to filter the data by
        :rtype: pandas DataFrame
        :return: dataframe of data from table
        """

        if gateways_list is not None:
            if len(gateways_list) == 1:
                gateways_list = list(gateways_list) + [""]
            gateways_list = tuple(gateways_list)
        # adding search by date to improve search time (most tables in data bricks partition is by date)
        query_start_datetime = convert_timestamp_to_datetime(str(start_time))
        query_end_datetime = convert_timestamp_to_datetime(str(end_time))
        start_datetime = query_start_datetime - datetime.timedelta(hours=36)
        end_datetime = query_end_datetime + datetime.timedelta(hours=36)
        start_date = datetime.datetime.strftime(start_datetime, '%Y%m%d')
        end_date = datetime.datetime.strftime(end_datetime, '%Y%m%d')
        sql_method = f"""
        select *
        from {table}
        where date between {start_date} and {end_date} 
        and timestamp between {start_time} and {end_time} 
        """
        if gateways_list is not None:
            sql_method = sql_method + f"""and gatewayId in {gateways_list}
        """
        if self.spark is not None:
            debug_print('Running SQL query...', center=True)
            debug_print(sql_method)
            query_data = self.spark.sql(sql_method)
            query_data = query_data.toPandas()
            return query_data
        else:
            raise EnvironmentError("SQL query can only run in databricks")

    def get_heartbeat_data(self, table, start_time, end_time, gateways_list=None, bridges_list=None):
        """
        does an SQL query of the statistics data table between specified timestamps
        :type table: str
        :param table: name of data table
        :type start_time: float
        :param start_time: time filter start timestamp (UTC milliseconds)
        :type end_time: float
        :param end_time: time filter end timestamp (UTC milliseconds)
        :type gateways_list: list
        :param gateways_list: list of gateways to filter the data by
        :rtype: pandas DataFrame
        :return: dataframe of data from table
        """

        if gateways_list is not None:
            if len(gateways_list) == 1:
                gateways_list = list(gateways_list) + [""]
            gateways_list = tuple(gateways_list)
        
        if bridges_list is not None:
            if len(bridges_list) == 1:
                bridges_list = list(bridges_list) + [""]
            bridges_list = tuple(bridges_list)
        
        # adding search by date to improve search time (most tables in data bricks partition is by date)
        query_start_datetime = convert_timestamp_to_datetime(str(start_time))
        query_end_datetime = convert_timestamp_to_datetime(str(end_time))
        start_datetime = query_start_datetime - datetime.timedelta(hours=36)
        end_datetime = query_end_datetime + datetime.timedelta(hours=36)
        start_date = datetime.datetime.strftime(start_datetime, '%Y%m%d')
        end_date = datetime.datetime.strftime(end_datetime, '%Y%m%d')
        sql_method = f"""
        select *
        from {table}
        where date between {start_date} and {end_date} 
        and timestamp between {start_time} and {end_time} 
        """
        if gateways_list is not None:
            sql_method = sql_method + f"and gatewayId in {gateways_list}\n"
        if bridges_list is not None:
            sql_method = sql_method + f"and bridgeId in {bridges_list}\n"
        if self.spark is not None:
            debug_print('Running SQL query...', center=True)
            debug_print(sql_method)
            query_data = self.spark.sql(sql_method)
            query_data = query_data.toPandas()
            return query_data
        else:
            raise EnvironmentError("SQL query can only run in databricks")

    
    def get_num_seen_bridges(self, table, start_time, end_time, gateways_list=None):
        """
        gets number of unique bridges seen by each gateways (bridge has to send data packets)
        :type table: str
        :param table: name of data table
        :type start_time: float
        :param start_time: time filter start timestamp (UTC milliseconds)
        :type end_time: float
        :param end_time: time filter end timestamp (UTC milliseconds)
        :type gateways_list: list
        :param gateways_list: list of gateways to filter the data by
        :return: dictionary of number of seen bridges per gateway
        """
        if gateways_list is not None:
            if len(gateways_list) == 1:
                gateways_list = list(gateways_list) + [""]
            gateways_list = tuple(gateways_list)
        # adding search by date to improve search time (most tables in data bricks partition is by date)
        query_start_datetime = convert_timestamp_to_datetime(str(start_time))
        query_end_datetime = convert_timestamp_to_datetime(str(end_time))
        start_datetime = query_start_datetime - datetime.timedelta(hours=36)
        end_datetime = query_end_datetime + datetime.timedelta(hours=36)
        start_date = datetime.datetime.strftime(start_datetime, '%Y%m%d')
        end_date = datetime.datetime.strftime(end_datetime, '%Y%m%d')
        sql_method = f"""
        select gatewayId, count(distinct bridgeId) as countBridge
        from {table}
        where date between {start_date} and {end_date} 
        and timestamp between {start_time} and {end_time} 
        """
        if gateways_list is not None:
            sql_method = sql_method + f"""and gatewayId in {gateways_list}
        """
        sql_method = sql_method + "group by gatewayId"
        if self.spark is not None:
            debug_print('Running SQL query...', center=True)
            debug_print(sql_method)
            query_data = self.spark.sql(sql_method)
            query_data = query_data.toPandas()
            query_data = dict(zip(query_data['gatewayId'], query_data['countBridge']))
            return query_data
        else:
            raise EnvironmentError("SQL query can only run in databricks")


    def get_sequence_id_loss(self, packets_table, statistics_table, start_time, end_time, gateways_list=None, sync_tables=True):
        """
        calculates sequence ID loss (takes reboot into consideration) per GW in table
        :type packets_table: str | DataFrame
        :param packets_table: name of data table | data table DataFrame
        :type statistics_table: str | DataFrame
        :param statistics_table: name of statistics table | statistics table DataFrame
        :type start_time: float
        :param start_time: time filter start timestamp (UTC milliseconds)
        :type end_time: float
        :param end_time: time filter end timestamp (UTC milliseconds)
        :type gateways_list: list
        :param gateways_list: list of gateways to filter the data by
        :type sync_tables: bool
        :param sync_tables: sync table start/end by timestamp (relevant for running directly after DB table updates)
        :return: dictionary of statistics
        """

        RESET_THRESHOLD = 1000
        
        def sync_data_and_statistics_tables(data_table, statistics_table, gw_list):
            """
            sync data and statistics table by timestamps
            """
            def get_first_stat_timestamp(statistics_table):
                first_stat_timestamp = 0
                for gw in statistics_table['gatewayId'].unique():
                    gw_first_stat_timestamp = np.sort(statistics_table[statistics_table['gatewayId'] == gw]['timestamp'])[0]
                    if gw_first_stat_timestamp > first_stat_timestamp:
                        first_stat_timestamp = gw_first_stat_timestamp
                return first_stat_timestamp
            
            # 'sync' statistics table and data table latest timestamps:
            max_data_timestamp = data_table['timestamp'].max()
            max_stats_timestamp = statistics_table['timestamp'].max()
            last_synced = min(max_data_timestamp, max_stats_timestamp)
            # truncate end of data tables to timeslot
            data_table = data_table[data_table['timestamp']<= last_synced]
            statistics_table = statistics_table[statistics_table['timestamp'] <= last_synced]
            
            first_stat_timestamp = get_first_stat_timestamp(statistics_table)
            stat_start_timestamp = first_stat_timestamp - (datetime.timedelta(seconds = 60).total_seconds() * 1000)
            data_start_timestamp = data_table['timestamp'].min()
            while data_start_timestamp > stat_start_timestamp:
                statistics_table = statistics_table[statistics_table['timestamp']>first_stat_timestamp]
                first_stat_timestamp = get_first_stat_timestamp(statistics_table)
                stat_start_timestamp = first_stat_timestamp - (datetime.timedelta(seconds = 60).total_seconds() * 1000)

            return data_table, statistics_table
                

        
        def get_stats_for_continuous_sequence_ids(sequence):
            """
            generate statistics for contiuous sequence IDs
            :type sequence: list
            :param sequence: array of sequence ID
            :rtype: dict
            :return: dictionary of statistics
            """
            stats = {}

            # remove duplicates
            sequence = np.unique(sequence)
            # sort by descending order
            sequence = -np.sort(-sequence)

            s_max = np.max(sequence)
            s_min = np.min(sequence)
            s_rec = len(np.unique(sequence)) # TODO - Compare with len(sequence) / num of coupled packets from table
            s_expected = (s_max - s_min)+1
            stats['maxSequenceId'] = s_max
            stats['minSequenceId'] = s_min
            stats['receivedSequenceIds'] = s_rec
            stats['expectedSequenceIds'] = s_expected
            return stats

        def process_sequences(df_array, num_mgmt_packets=0):
            """
            generate statistics for (normalized by duration) for array of sequence IDs (compensated for GW Reboots)
            :type df_array: list of DataFrames
            :param df_array: list of DataFrames with continuous sequence IDs
            :type num_mgmt_packets: int
            :param num_mgmt_packets: number of management packets
            :rtype: dict
            :return: dictionary of statistics
            """
            total_stats = {}
            total_received_packets = num_mgmt_packets
            total_expected_packets = 0
            for df in df_array:
                stats = get_stats_for_continuous_sequence_ids(df['sequenceId'])
                total_received_packets += stats['receivedSequenceIds']
                total_expected_packets += stats['expectedSequenceIds']
            if total_expected_packets == 0:
                breakpoint()
            loss_percentage = (1 - (total_received_packets / total_expected_packets) )* 100
            total_stats['totalManagementPackets'] = num_mgmt_packets
            total_stats['totalDataPackets'] = total_received_packets - num_mgmt_packets
            total_stats['totalReceivedPackets'] = total_received_packets
            total_stats['totalExpectedPackets'] = total_expected_packets
            total_stats['lossPercentage'] = round(loss_percentage, 3)
            total_stats['numResets'] = len(df_array)-1
            return total_stats

        results = {}
        if isinstance(packets_table, pd.DataFrame):
            data = packets_table
        else:
            data = self.get_sequence_id_data(packets_table, start_time, end_time, gateways_list).sort_values(by='timestamp', ascending=False)
        gw_list = data['gatewayId'].unique()
        if gw_list is None:
            return None
        if isinstance(statistics_table, pd.DataFrame):
            statistics_data = statistics_table
        else:
            statistics_data = self.get_statistics_data(statistics_table, start_time, end_time, gw_list)
        if sync_tables:
            data, statistics_data = sync_data_and_statistics_tables(data, statistics_data, gw_list)

        for gw in gw_list:
            gw_name = data[data['gatewayId'] == gw]['gatewayName'].iloc[0]
            num_mgmt_packets = statistics_data[statistics_data['gatewayId'] == gw]['managementPktCount'].sum()
            gw_results = []
            gw_df = data[data['gatewayId'] == gw].reset_index()
            gw_df['diff'] = gw_df['sequenceId'].diff()
            gw_df['reset'] = np.where(gw_df['diff']>RESET_THRESHOLD, True, False)
            gw_resets = gw_df[gw_df['reset'] == True]
            sequences = np.array_split(gw_df, gw_resets.index)
            gw_results = process_sequences(sequences, num_mgmt_packets)
            gw_results.update({'gwName': gw_name})
            results[gw] = gw_results
        return results



    def get_amount_of_unique_tags_per_data_path(self, table, start_time, end_time, gateways_list=None,
                                                tags_to_ignore=None, bridges_list=None):
        """
        does an SQL query of the packet data table between specified timestamps
        returns amount of unique externalIds per data path (bridge->gw) in the given timeframe
        :type table: str
        :param table: name of data table
        :type start_time: float
        :param start_time: time filter start timestamp (UTC milliseconds)
        :type end_time: float
        :param end_time: time filter end timestamp (UTC milliseconds)
        :type tags_to_ignore: list
        :param tags_to_ignore: list of tags to ignore in the query (will not be counted)
        :type gateways_list: list
        :param gateways_list: list of gateways to filter the data by
        :type bridges_list: list
        :param bridges_list: list of bridges to filter the data by
        :rtype: pandas DataFrame
        :return: dataframe of data from table
        """
        if bridges_list is not None:
            if len(bridges_list) == 1:
                bridges_list = list(bridges_list) + [""]
            bridges_list = tuple(bridges_list)
        if gateways_list is not None:
            if len(gateways_list) == 1:
                gateways_list = list(gateways_list) + [""]
            gateways_list = tuple(gateways_list)
        if tags_to_ignore is not None:
            if len(tags_to_ignore) == 1:
                tags_to_ignore = list(tags_to_ignore) + [""]
            tags_to_ignore = tuple(tags_to_ignore)
        # adding search by date to improve search time (most tables in data bricks partition is by date)
        query_start_datetime = convert_timestamp_to_datetime(str(start_time))
        query_end_datetime = convert_timestamp_to_datetime(str(end_time))
        start_datetime = query_start_datetime - datetime.timedelta(hours=36)
        end_datetime = query_end_datetime + datetime.timedelta(hours=36)
        start_date = datetime.datetime.strftime(start_datetime, '%Y%m%d')
        end_date = datetime.datetime.strftime(end_datetime, '%Y%m%d')
        sql_method = f"""
        select gatewayId, bridgeId, count(distinct externalId)
        from {table}
        where date between {start_date} and {end_date} 
        and timestamp between {start_time} and {end_time} 
        """
        if tags_to_ignore is not None and tags_to_ignore != ():
            sql_method = sql_method + f"""and externalId not in {tags_to_ignore}
        """
        if bridges_list is not None and bridges_list != ():
            sql_method = sql_method + f"""and bridgeId in {bridges_list}
        """
        if gateways_list is not None and gateways_list != ():
            sql_method = sql_method + f"""and gatewayId in {gateways_list}
        """
        sql_method = sql_method + f"""group by gatewayId, bridgeId   order by gatewayId, bridgeId"""
        if self.spark is not None:
            debug_print('Running SQL query...', center=True)
            debug_print(sql_method)
            query_data = self.spark.sql(sql_method)
            query_data = query_data.toPandas()
            return query_data
        else:
            raise EnvironmentError("SQL query can only run in databricks")

    def get_amount_of_unique_tags_per_brg(self, table, start_time, end_time, tags_to_ignore=None, bridges_list=None):
        """
        does an SQL query of the packet data table between specified timestamps
        returns amount of unique externalIds per data path (bridge->gw) in the given timeframe
        :type table: str
        :param table: name of data table
        :type start_time: float
        :param start_time: time filter start timestamp (UTC milliseconds)
        :type end_time: float
        :param end_time: time filter end timestamp (UTC milliseconds)
        :type tags_to_ignore: list
        :param tags_to_ignore: list of tags to ignore in the query (will not be counted)
        :type gateways_list: list
        :param gateways_list: list of gateways to filter the data by
        :type bridges_list: list
        :param bridges_list: list of bridges to filter the data by
        :rtype: pandas DataFrame
        :return: dataframe of data from table
        """
        if bridges_list is not None:
            if len(bridges_list) == 1:
                bridges_list = list(bridges_list) + [""]
            bridges_list = tuple(bridges_list)
        if tags_to_ignore is not None:
            if len(tags_to_ignore) == 1:
                tags_to_ignore = list(tags_to_ignore) + [""]
            tags_to_ignore = tuple(tags_to_ignore)
        # adding search by date to improve search time (most tables in data bricks partition is by date)
        query_start_datetime = convert_timestamp_to_datetime(str(start_time))
        query_end_datetime = convert_timestamp_to_datetime(str(end_time))
        start_datetime = query_start_datetime - datetime.timedelta(hours=36)
        end_datetime = query_end_datetime + datetime.timedelta(hours=36)
        start_date = datetime.datetime.strftime(start_datetime, '%Y%m%d')
        end_date = datetime.datetime.strftime(end_datetime, '%Y%m%d')
        sql_method = f"""
        select gatewayId, bridgeId, count(distinct externalId)
        from {table}
        where date between {start_date} and {end_date} 
        and timestamp between {start_time} and {end_time} 
        """
        if tags_to_ignore is not None and tags_to_ignore != ():
            sql_method = sql_method + f"""and externalId not in {tags_to_ignore}
        """
        if bridges_list is not None and bridges_list != ():
            sql_method = sql_method + f"""and bridgeId in {bridges_list}
        """
        sql_method = sql_method + f"""group by bridgeId   order by bridgeId"""
        if self.spark is not None:
            debug_print('Running SQL query...', center=True)
            debug_print(sql_method)
            query_data = self.spark.sql(sql_method)
            query_data = query_data.toPandas()
            return query_data
        else:
            raise EnvironmentError("SQL query can only run in databricks")

    def get_amount_of_unique_tags_per_gw(self, table, start_time, end_time, gateways_list=None, tags_to_ignore=None):
        """
        does an SQL query of the packet data table between specified timestamps
        returns amount of unique externalIds per data path (bridge->gw) in the given timeframe
        :type table: str
        :param table: name of data table
        :type start_time: float
        :param start_time: time filter start timestamp (UTC milliseconds)
        :type end_time: float
        :param end_time: time filter end timestamp (UTC milliseconds)
        :type tags_to_ignore: list
        :param tags_to_ignore: list of tags to ignore in the query (will not be counted)
        :type gateways_list: list
        :param gateways_list: list of gateways to filter the data by
        :rtype: pandas DataFrame
        :return: dataframe of data from table
        """

        if gateways_list is not None:
            if len(gateways_list) == 1:
                gateways_list = list(gateways_list) + [""]
            gateways_list = tuple(gateways_list)
        if tags_to_ignore is not None:
            if len(tags_to_ignore) == 1:
                tags_to_ignore = list(tags_to_ignore) + [""]
            tags_to_ignore = tuple(tags_to_ignore)
        # adding search by date to improve search time (most tables in data bricks partition is by date)
        query_start_datetime = convert_timestamp_to_datetime(str(start_time))
        query_end_datetime = convert_timestamp_to_datetime(str(end_time))
        start_datetime = query_start_datetime - datetime.timedelta(hours=36)
        end_datetime = query_end_datetime + datetime.timedelta(hours=36)
        start_date = datetime.datetime.strftime(start_datetime, '%Y%m%d')
        end_date = datetime.datetime.strftime(end_datetime, '%Y%m%d')
        sql_method = f"""
        select gatewayId, count(distinct externalId)
        from {table}
        where date between {start_date} and {end_date} 
        and timestamp between {start_time} and {end_time} 
        """
        if tags_to_ignore is not None and tags_to_ignore != ():
            sql_method = sql_method + f"""and externalId not in {tags_to_ignore}
        """
        if gateways_list is not None and gateways_list != ():
            sql_method = sql_method + f"""and gatewayId in {gateways_list}
        """
        sql_method = sql_method + f"""group by gatewayId   order by gatewayId"""
        if self.spark is not None:
            debug_print('Running SQL query...', center=True)
            debug_print(sql_method)
            query_data = self.spark.sql(sql_method)
            query_data = query_data.toPandas()
            return query_data
        else:
            raise EnvironmentError("SQL query can only run in databricks")

    def get_last_data(self, table, mins=60, tags_list=None, bridges_list=None, gateways_list=None):
        """
        function querys back specified minutes from data table (counts back from last entry in table)
        :type table: str
        :param table: databricks data table name
        :type mins: int
        :param mins: minutes to query back, defaults to 1 hour
        :param tags_list: tags list
        :type tags_list: list
        :param tags_list: list of tags to filter from the data
        :type bridges_list: list
        :param bridges_list: list of bridger to filter from the data
        :type gateways_list: list
        :param gateways_list: list of gateways to filter from the data
        :rtype: pandas DataFrame
        :return: dataframe of data from table
        """
        end_timestamp = self.get_last_entry_timestamp(table)
        end_datetime = mstimestamp_to_timezone(end_timestamp)
        start_timestamp = end_timestamp - (mins * 60 * 1000)
        start_datetime = mstimestamp_to_timezone(start_timestamp)
        debug_print(f'Getting last {mins} mins from {table}')
        debug_print(f'Last entry at {end_timestamp}')
        debug_print(f'querying {start_datetime} -> {end_datetime}')
        return self.get_seen_tags(table, start_timestamp, end_timestamp, tags_list, bridges_list, gateways_list)

    def get_last_entry_timestamp(self, table):
        """
        function gets name of data table and returns the timestamp of the last entry (in milliseconds)
        :type table: str
        :param table: name of table
        """
        today = datetime.datetime.now()
        yesterday = today-datetime.timedelta(days=1)
        today_date = datetime.datetime.strftime(today, '%Y%m%d')
        yesterday_date = datetime.datetime.strftime(yesterday, '%Y%m%d')

        # Try to get last timestamp from last day
        sql_method = f"""
        select MAX (timestamp)
        from {table}
        where date between {yesterday_date} and {today_date} 
        """
        if self.spark is not None:
            query_data = self.spark.sql(sql_method)
            query_data = query_data.toPandas()
            last_entry_ts = query_data.iloc[0][0]
            if not np.isnan(last_entry_ts):
                return query_data.iloc[0][0]
            else:
                # Query all packets in table
                sql_method = ''.join(sql_method.split('\n')[:-2])
                query_data = self.spark.sql(sql_method)
                query_data = query_data.toPandas()
                last_entry_ts = query_data.iloc[0][0]
                if not np.isnan(last_entry_ts):
                    return query_data.iloc[0][0]
                else:
                    raise WiliotTableError(f'Cannot get last entry, no entries in data table {table}')
        else:
            raise EnvironmentError("Unable to detect dbutils function")

    def wait_for_data(self, table, requested_timestamp, timeout_mins=70):
    # TODO - implement date search
        """
        function waits for data timed at requested_timestamp (or later) to appear in the data table.
        :type table: str
        :param table: data table name
        :type requested_timestamp: int
        :param requested_timestamp: timestamp to wait to appear in the table
        """
        last_entry_timestamp = self.get_last_entry_timestamp(table)
        table_ready = last_entry_timestamp >= requested_timestamp
        start = current_timestamp()
        while not table_ready:
            if (current_timestamp() - start) >= 1000 * 60 * timeout_mins:
                debug_print(f"""{timeout_mins} minute timeout reached! Table still does not have the requested data
                Run test again or get raw data with wait_for_data=False""")
                exit_notebook(f'Wait for table timeout reached after {timeout_mins} minutes')
            debug_print(
                f'Waiting for table to get data. \n'
                f'Requested timestamp {requested_timestamp} | ({mstimestamp_to_timezone(requested_timestamp)}) \n'
                f'Table latest timestamp {last_entry_timestamp} | ({mstimestamp_to_timezone(last_entry_timestamp)})')
            mins_behind = (requested_timestamp - last_entry_timestamp) / 60000
            debug_print(f'Data table {mins_behind} minutes behind , trying again in 30 seconds', center=True)
            sleep(30)
            last_entry_timestamp = self.get_last_entry_timestamp(table)
            table_ready = last_entry_timestamp >= requested_timestamp
        debug_print(
            f'Table is updated. \n'
            f'Requested timestamp {requested_timestamp}, table latest timestamp {last_entry_timestamp} \n'
            f'Table latest timestamp {last_entry_timestamp} | ({mstimestamp_to_timezone(last_entry_timestamp)})')
        mins_ahead = (last_entry_timestamp - requested_timestamp) / 60000
        debug_print(f'Data table {mins_ahead} minutes ahead', center=True)

    def get_seen_edge_devices_from_packets(self, table, start_time, end_time):
        """
        does an SQL query of the packet data table between specified timestamps
        returns dictionary of seen gateways and bridges from slice of packet table
        :type table: str
        :param table: name of data table
        :type start_time: float
        :param start_time: time filter start timestamp (UTC milliseconds)
        :type end_time: float
        :param end_time: time filter end timestamp (UTC milliseconds)
        :rtype: pandas DataFrame
        :return: dataframe of data from table
        """
        # adding search by date to improve search time (most tables in data bricks partition is by date)
        query_start_datetime = convert_timestamp_to_datetime(str(start_time))
        query_end_datetime = convert_timestamp_to_datetime(str(end_time))
        start_datetime = query_start_datetime - datetime.timedelta(hours=24)
        end_datetime = query_end_datetime + datetime.timedelta(hours=24)
        start_date = datetime.datetime.strftime(start_datetime, '%Y%m%d')
        end_date = datetime.datetime.strftime(end_datetime, '%Y%m%d')
        sql_method = f"""
        select gatewayId, bridgeId, gatewayName, max(timestamp) as timestamp
        from {table}
        where date between {start_date} and {end_date} 
        and timestamp between {start_time} and {end_time}
        group by gatewayId, gatewayName, bridgeId 
        """
        if self.spark is not None:
            debug_print('Running SQL query...', center=True)
            debug_print(sql_method)
            query_data = self.spark.sql(sql_method).toPandas()
            query_data['gatewayId'] = query_data['gatewayId'].str.upper()
            query_data['bridgeId'] = query_data['bridgeId'].str.upper()
            return query_data
        else:
            raise EnvironmentError("SQL query can only run in databricks")


def process_tagstats(params, tag_stats_df, working_directory, test_no):
    """
        function gets tag stats dataframe and creates downloadable cdf graph for specified parameters
        :type params: list
        :param params: tag stats metrics to create graphs for
        :type tag_stats_df: pandas DataFrame
        :param tag_stats_df: tag stats DataFrame (can be created using get_tagstats_from_test)
        :type working_directory: string
        :param working_directory: directory to save the graphs HTML files
        :type test_no: string
        :param test_no: test number
    """
    if params is None:
        debug_print('No parameters specified. Creating graphs for all possible parameters.')
        params = ['num_packets', 'num_cycles', 'sprinkler_counter_mean', 'sprinkler_counter_std',
                  'sprinkler_counter_min', 'sprinkler_counter_max', 'tbp_mean', 'tbp_std', 'tbp_min', 'tbp_max',
                  'tbp_num_vals', 'per_mean', 'per_std', 'rssi_mean', 'rssi_std', 'rssi_min', 'rssi_max', 'ttfp',
                  'ttfp_seconds', 'end_time', 'duration', 'rx_rate_normalized', 'rx_rate', 'charge_time_min',
                  'charge_time_max', 'packet_counter_min', 'packet_counter_max', 'packet_counter_first',
                  'estimated_packet_counter_resets', 'estimated_total_packet_rate', 'estimated_total_per',
                  'externalId']
    tests_list = tag_stats_df['testId'].unique()
    for param in params:
        debug_print(f"""*****************************{param}*****************************""")
        graph = pd.DataFrame()
        for test in tests_list:
            df = tag_stats_df[tag_stats_df.testId == test]
            test_curveset = cdf_dataframe(df, param, test)
            graph = pd.concat([test_curveset, graph])
            debug_print(f"""test {test} added to graph""")
        fig = px.scatter(graph, x=param, y='TagCount', color=['testName'])
        fig.update_layout(title=param)
        fig.update_yaxes()
        if param == 'ttfp_seconds':
            fig.update_xaxes(title_text='Time to first packet [seconds]')
        if param == 'estimated_total_packet_rate':
            fig.update_xaxes(title_text='Estimated total packet rate [counters/second]')
        filepath = f"""{working_directory}{test_no}_{param}_graph.html"""
        open(filepath, 'a').close()
        fig.write_html(filepath)
        # debug_print(filepath)
        # p = plot(fig, output_type='div')
        # display_html(p)
        create_download_link(filepath, file_name=f"""{param} graph""")

def file_exists(path):
    """
    function checks if file exists
    :type path: str
    :param path: path to file
    :rtype: bool
    :return: file exists
    """
    path = path.replace('/dbfs', 'dbfs:')
    try:
        db_utils().fs.head(path)
        return True
    except Exception:
        try:
            return exists(path)
        except Exception:
            return False

def get_tagstats_from_test(tests_config_df=None, tests_config_csv_path=None, rawdata_df=None,
                           rawdata_csv_path=None, tagdeployment_df=None, tagdeployment_csv_path=None):
    """
        function gets test config df/csv and rawdata (from datatables) df/csv as input
        the function processes the data and created a new dataframe containing tag statistics for each test.
        :type tests_config_df: pandas DataFrame
        :param tests_config_df: dataframe of test config
        :type tests_config_csv_path: string
        :param tests_config_csv_path: path to test config csv
        :type rawdata_df: pandas DataFrame
        :param rawdata_df: dataframe of raw data
        :type rawdata_csv_path: string
        :param rawdata_csv_path: path to raw data csv
        :type tagdeployment_df: pandas DataFrame
        :param tagdeployment_df:dataframe of tag deployment data
        :type tagdeployment_csv_path: string
        :param tagdeployment_csv_path: path to tag deployment data csv
    """
    if tests_config_csv_path is not None and not tests_config_df:
        tests_config_df = pd.read_csv(tests_config_csv_path)
    if rawdata_csv_path is not None and not rawdata_df:
        rawdata_df = pd.read_csv(rawdata_csv_path)
    if tagdeployment_csv_path is not None:
        tagdeployment_df = pd.read_csv(tagdeployment_csv_path)
    testid_list = tests_config_df['testId'].unique()
    master_tag_stats = pd.DataFrame()
    for testid in testid_list:
        if  testid not in rawdata_df.testId.unique():
            debug_print(f"""0 data for testId {testid} in raw data! please check tables/test config""")
            continue
        currenttest_df = rawdata_df[rawdata_df.testId == testid].copy()
        start_time = tests_config_df[tests_config_df.testId == testid].get('startTimestamp').item()
        debug_print(f"""***** Processing {testid} data. Test Timestart = {start_time}""")
        currenttest_packets = analyze_packets_data(packets_df=currenttest_df, test_time_start=start_time)
        tag_stats_dict = currenttest_packets.get_group_statistics(group_by_col='externalId')
        tag_stats_df = pd.DataFrame.from_dict(tag_stats_dict, orient='index')
        tag_stats_df['testId'] = testid
        master_tag_stats = pd.concat([master_tag_stats, tag_stats_df])
    if tagdeployment_df is not None:
        debug_print('***********Tag deployment information merged with tag stats from test***********')
        master_tag_stats = pd.merge(master_tag_stats, tagdeployment_df, how='outer', on='externalId'). \
            sort_values('testId').set_index('externalId')
    master_tag_stats.insert(18, 'ttfp_seconds', master_tag_stats.ttfp / 1000)
    return master_tag_stats

def massagedata(dataset, param):
    curveset = pd.DataFrame(columns=[param, 'TagCount'])
    count = 0
    reftme_stmp = None
    for index, row in dataset.iterrows():
        if pd.isna(row[param]):
            continue
        if reftme_stmp is None:
            reftme_stmp = row[param]
        if reftme_stmp == row[param]:
            count += 1
        else:
            curveset = curveset.append({param: reftme_stmp, 'TagCount': count}, ignore_index=True)
            reftme_stmp = row[param]
            count += 1
    curveset = curveset.append({param: reftme_stmp, 'TagCount': count}, ignore_index=True)
    return curveset

def display_html(html):
    """
    Use databricks displayHTML from an external package
    :type html: string
    :param html: html document to display
    """
    for frame in inspect.getouterframes(inspect.currentframe()):
        global_names = set(frame.frame.f_globals)
        # Use multiple functions to reduce risk of mismatch
        if all(v in global_names for v in ["displayHTML", "display", "spark"]):
            return frame.frame.f_globals["displayHTML"](html)
    raise EnvironmentError("Unable to detect displayHTML function")

def display_print(todisplay, console_only=False, **kwargs):
    """
    Use databricks display func from an external package.
    uses tabulate library to display dataframe in case running locally
    :type todisplay: pandas DataFrame
    :param todisplay: variable to display
    :type console_only: bool
    :param console_only: if true, prints the data table to console (even if running in DB notebook) using tabulate
    """
    if not console_only:
        for frame in inspect.getouterframes(inspect.currentframe()):
            # call dbutils display
            global_names = set(frame.frame.f_globals)
            if all(v in global_names for v in ["display"]):
                try:
                    return frame.frame.f_globals["display"](todisplay)
                except ValueError:
                    debug_print('ValueError when reading DataFrame! Trying to print in console', center=True)
                except KeyError:
                    debug_print('KeyError when reading DataFrame! Trying to print in console', center=True)
                except TypeError:
                    debug_print('TypeError when reading DataFrame! Trying to print in console', center=True)
    if isinstance(todisplay, pd.DataFrame) or console_only:
        debug_print('\n' + tabulate.tabulate(todisplay, **kwargs))
        return None
    raise EnvironmentError("Unable to detect Display function")

def db_utils_rm(dbfs_path):
    """
    helper function to remove files/folders from dbfs
    """
    dbfs_path = dbfs_path.replace('/dbfs', 'dbfs:')
    return db_utils().fs.rm(dbfs_path)

def db_utils():
    """
    Use databricks dbutils from an external package
    for example: to use dbutils.fs.head() - use db_utils().fs.head()
    """
    for frame in inspect.getouterframes(inspect.currentframe()):
        global_names = set(frame.frame.f_globals)
        # Use multiple functions to reduce risk of mismatch
        if all(v in global_names for v in ["dbutils"]):
            return frame.frame.f_globals["dbutils"]()
    raise EnvironmentError("Unable to detect dbutils function")

def get_secret(scope, key):
    """
    get databrickst secret, return None if not running in databricks
    :param scope: secret scope
    :param key: secret key
    """
    try:
        secret = db_utils().secrets.get(scope=scope, key=key)
        return secret
    except EnvironmentError:
        raise(EnvironmentError('Cannot get secret when not running in databricks!'))


def create_download_link(dbfs_path, file_name='file'):
    """
    accepts path to dbfs file, and creates a download link to the file in the notebook
    the function only works with files saved in /dbfs/FileStore
    :type dbfs_path: string
    :param dbfs_path: path to dbfs path (accepts either spark API or file API format)
    :type file_name: string
    :param file_name: name of file (this will be the name of the link created)
    """
    if is_databricks():
        if not ('/dbfs/FileStore' in dbfs_path or 'dbfs:/FileStore' in dbfs_path):
            raise ValueError('the path must start with /dbfs/FileStore or dbfs:/FileStore!')
        dbfs_path = dbfs_path.replace('/dbfs/FileStore', '/files')
        dbfs_path = dbfs_path.replace('dbfs:/FileStore', '/files')
        display_html(f"""<a href="{dbfs_path}" download>Download {file_name} </a>""")
    debug_print(f"""File available at {dbfs_path}""")

def cdf_dataframe(dataset, param, dataset_name, groupby=None, comp=False):
    """
    function accepts dataframe as input the function returns dataframe of CDF
    complementary cumulative distribution function (of tag count) according to param,
    and labels it by dataset_name
    :type dataset: pandas DataFrame
    :param dataset: dataset to create CDF
    :type param: string
    :param param: parameter to create the CDF
    :type dataset_name: string
    :param dataset_name: name to add to the result
    :type groupby: str
    :param groupby: added parameter (in dataset) to group results by
    :type comp: bool
    :param comp: whether to create CDF (comp = False) or CCDF (comp = True)
    """
    curveset = pd.DataFrame(columns=[param, 'TagCount', groupby])
    dataset = dataset.sort_values(param, ascending=True)
    if groupby is None:
        iterations = [None]
    else:
        iterations = dataset[groupby].unique()
    for group in iterations:
        group_curveset = pd.DataFrame(columns=[param, 'TagCount', groupby])
        if group is not None:
            group_dataset = dataset[dataset[groupby] == group]
        else:
            group_dataset = dataset
        values = group_dataset[param].unique()
        group_curveset[param] = values
        if group is not None:
            group_curveset[groupby] = group
        for index, value in group_curveset[param].items():
            biggerthan = group_dataset[group_dataset[param] > value][param].count()
            smallerorequals = group_dataset[group_dataset[param] <= value][param].count()
            if comp:
                group_curveset.at[index, 'TagCount'] = biggerthan
            else:
                group_curveset.at[index, 'TagCount'] = smallerorequals
        if comp:
            group_curveset['testName'] = dataset_name + '_CCDF'
        else:
            group_curveset['testName'] = dataset_name + '_CDF'
        curveset = pd.concat([curveset, group_curveset])

    return curveset

def powerset(iterable):
    s = list(iterable)
    return chain.from_iterable(combinations(s, r) for r in range(len(s) + 1))

def analyze_packets_data(packets_df=None, csv_path=None, test_time_start=None, test_time_conf_received=None,
                         no_duplications=True):

    """
    gets dataframe of packets / csv of packets, returns NetworkPacketList of packets
    :type csv_path: string
    :param csv_path: path to the packets csv file
    :type packets_df: dataframe
    :param packets_df: packets dataframe
    :type test_time_start: timestamp (int)
    :param test_time_start: UTC millisecond timestamp of the beginning of the test
    :type test_time_conf_received: timestamp (int)
    :param test_time_conf_received: UTC millisecond timestamp of the time the test configuration was received
    :type no_duplications: bool
    :param no_duplications: if False, will keep duplications of same rawPacket
    :return: NetworkPacketList created from .csv located in path
    """
    if test_time_start is None:
        debug_print('Test time start was not passed, some metrics will not be calculated!')
    else:
        debug_print(f'Test time start {test_time_start} | ')
    if csv_path is None and packets_df is None:
        debug_print("no data was given!()")
        return None
    if csv_path is not None:
        packets_df = pd.read_csv(csv_path)

    debug_print('Preprocessing raw data...')

    # remove duplicate rows
    dup_rows = packets_df.shape[0]
    packets_df.sort_values(by='timestamp', ascending=True, ignore_index=True, inplace=True)
    if no_duplications:
        packets_df.drop_duplicates('rawPacket', ignore_index=True, inplace=True)
        dup_rows = dup_rows - packets_df.shape[0]
        if dup_rows > 0:
            debug_print(f'{dup_rows} duplicate rows removed')

    # fix adva
    packets_df['rawPacket'] = packets_df['rawPacket'].apply(lambda x: x[4:] if x.startswith('1E16') else x)

    # flag bugged tags packets
    flag_bugged_tags(packets_df)

    debug_print('Creating NetworkPacketList from DataFrame...')
    packets_list = NetworkPacketList(list_custom_data={'test_time_start': test_time_start})
    if test_time_conf_received is not None:
        packets_list.list_custom_data['test_time_conf_received'] = test_time_conf_received
    reconstructed_packet_df = network_data_preparation(packets_df=packets_df)

    packets_list = packets_list.import_packet_df(import_all=True, packet_df=reconstructed_packet_df)
    return packets_list

def create_zip(dir_path):
    """
    gets path to FileStore directory and creates a download link to a zip file of the files in given directory
    zip file will be saved inside the dir_path directory
    :type dir_path: string
    :param dir_path: path to FileStore directory to zip
    """
    if dir_path[-1] == '/':
        dir_path = dir_path[:-1]
    dir_name = dir_path.rsplit('/', 1)[-1]
    if file_exists(dir_path + '/' + dir_name + '.zip'):
        db_utils().fs.rm(dir_path.replace('/dbfs', 'dbfs:') + '/' + dir_name + '.zip')
    zip_tmp_path = shutil.make_archive('/tmp/zips/tmpzip', 'zip', dir_path)
    zip_download_path = dir_path.replace('/dbfs', 'dbfs:') + '/' + dir_name + '.zip'
    db_utils().fs.mv(f'file:{zip_tmp_path}', zip_download_path)
    create_download_link(zip_download_path, f'{dir_name}.zip')

def exit_notebook(message):
    """
    closes notebook and prints out error message
    :type message: str
    :param message: error message
    """
    debug_print(message, center=True)
    try:
        db_utils().notebook.exit(message)
    except EnvironmentError as e:
        raise Exception(message)

def save_df(df, path, name=None, withindex=False, silent=False):
    """
    saves DataFrame to path, displays the DataFrame and creates a download link
    :type df: pandas Dataframe
    :param df: dataframe to save
    :type path: str
    :param path: path to save can be entered either as filename to save or directory to save in
    :type name: str
    :param name: name of dataframe (this displays in the download link and as a header before the link
    :type withindex: bool
    :param withindex: flag to choose if to export the dataframe with/without index
    :type silent: bool
    :param silent: if true does not generate HTML link
    """
    # save as dataframe.csv if no filename given
    if path[-1] == '/':
        path = path + 'dataframe'
    if path[-4:] != '.csv':
        path = path + '.csv'
    if name is None:
        name = path.split()[-1]
    if is_databricks():
        path = path.replace(' ', '_')
    debug_print(f'Saving DataFrame at {path}')
    if not is_databricks():
        df.to_csv(path, index=withindex)
        return True
    try:
        df.to_csv(path, index=withindex)
    except OSError:
        mkdirs_path = path.replace('/dbfs', 'dbfs:').rsplit('/', 1)[0]
        db_utils().fs.mkdirs(mkdirs_path)
        df.to_csv(path, index=withindex)
    if not silent:
        if is_databricks():
            display_html(f'<h1>{name}</h1>')
            create_download_link(dbfs_path=path, file_name=name)
            display_print(pd.read_csv(path), headers="keys")
        else:
            debug_print(f'{name} available at {path}')

def initialize_logger(working_directory=None):
    """
    initializes the logger to print to log and to logfile, which by default is named by the current timestamp (ms)
    when calling the function.
    :param working_directory: working directory to save logfile (in case running locally)
    :type working_directory: str
    :return: logger fileHandler filename
    """
    logging.shutdown()
    reload(logging)
    logger = logging.getLogger()
    for handler in logger.handlers:
        try:
            # extract current filename from log
            filename = handler.baseFilename.split('\\')[-1].split('.')[0]
            debug_print(f'Logger already initialized! passing logfile {filename}')
            return filename
        except Exception:
            pass
    logger_filename = int(current_timestamp())
    if is_databricks():
        db_utils().fs.put(f'file:/databricks/driver/{logger_filename}.log', '', overwrite=True)
        logging.getLogger("py4j.java_gateway").setLevel(logging.ERROR)
        logging.basicConfig(level=logging.DEBUG,
                            format='%(asctime)s | %(levelname)s | %(message)s',
                            handlers=[
                                logging.FileHandler(f'{logger_filename}.log', 'a'),
                                # logging.handlers.RotatingFileHandler
                                logging.StreamHandler()
                            ], force=True)
        logging.getLogger().handlers[0].setLevel(logging.DEBUG)
        logging.getLogger().handlers[1].setLevel(logging.INFO)
        debug_print(f'logger initialized at {logger_filename}', center=True)
        debug_print(f'logfile located at file:/databricks/driver/{logger_filename}.log')
    else:
        if working_directory is None:
            working_directory = os.path.join(user_data_dir(), 'wiliot', 'deployment_tools')
        if not os.path.exists(working_directory):
            os.makedirs(working_directory)
        logging.basicConfig(level=logging.DEBUG,
                            format='%(asctime)s | %(levelname)s | %(message)s',
                            handlers=[
                                logging.FileHandler(f'{working_directory}/{logger_filename}.log', 'a'),
                                # logging.handlers.RotatingFileHandler
                                logging.StreamHandler()
                            ])
        # filter stream to show info and up
        logging.getLogger().handlers[0].setLevel(logging.DEBUG)
        logging.getLogger().handlers[1].setLevel(logging.INFO)
        debug_print(f'logger initialized at {logger_filename}', center=True)
        debug_print(f'logfile located at {working_directory}/{logger_filename}.log')
    logging.getLogger().setLevel(logging.DEBUG)
    return logger_filename

def create_logfile(logger_filename, working_directory, copy_filename):
    """
        function copies the logfile for recent run to the test working directory and creates link for download
        :type logger_filename: str
        :param logger_filename: filename of log file
        :type working_directory: str
        :param working_directory: directory to copy logfile to
        :type copy_filename: str
        :param copy_filename: name of logfile when copied
    """
    if not is_databricks():
        debug_print('Cannot create logfile, logfile available already!')
    debug_print('Creating Logfile...', center=True)
    debug_print(f'cp {logger_filename}.log {copy_filename}.log')
    subprocess.run(f'cp {logger_filename}.log {copy_filename}.log', shell=True)
    copy_directory = working_directory.replace('/dbfs', 'dbfs:')
    debug_print(f'copy directory {copy_directory}')
    db_utils().fs.mkdirs(f'{copy_directory}')
    debug_print(f'mkdirs {copy_directory}')
    db_utils().fs.cp(f'file:/databricks/driver/{copy_filename}.log', copy_directory)
    debug_print(f'cp file:/databricks/driver/{copy_filename}.log, copy_directory')
    create_download_link(working_directory + f'{copy_filename}.log', f'{copy_filename}.log')
    debug_print(f"create download link - {working_directory} + f'{copy_filename}.log', f'{copy_filename}.log'")

def get_packet_table_name(owner, env, platform=False):
    """
    function gets ownerId and environment and return the name of the packet table (in databricks)
    :type owner: str
    :param owner: ownerId
    :type env: str
    :param env: wiliot environment (prod/test/dev)
    :type platform: bool
    :param platform: wiliot platform
    :rtype: str
    :return: data table name
    """
    env = 'prod' if env is None else env
    if not platform:
        data_table = owner + '' + f'_packet_data_{env}'
    else:
        data_table = owner + f'_enriched_packets_data_{env}'
    if owner.isnumeric():
        data_table = '_' + data_table
    return data_table

def get_event_table_name(owner, env=None, platform=False):
    """
    function gets ownerId and environment and return the name of the event table (in databricks)
    :type owner: str
    :param owner: ownerId
    :type env: str
    :param env: wiliot environment (prod/test/dev)
    :type platform: bool
    :param platform: wiliot platform
    :rtype: str
    :return: data table name
    """
    env = 'prod' if env is None else env
    if not platform:
        event_table = owner + '' + f'_event_data_{env}'
    else:
        event_table = owner + '' + f'_assets_metrics_data_{env}'
    if owner.isnumeric():
        event_table = '_' + event_table
    return event_table

def get_heartbeat_table_name(env=None):
    """
    function gets environment and returns the name of the heartbeat table (in databricks)
    :type owner: str
    :param owner: ownerId
    :type env: str
    :param env: wiliot environment (prod/test/dev)
    :rtype: str
    :return: hearbeat table name
    """
    env = 'prod' if env is None else env
    hearbeat_table = f'_network_heartbeat_statistics_{env}'
    return hearbeat_table

def get_statistics_table_name(env=None):
    """
    function gets environment and returns the name of the heartbeat table (in databricks)
    :type owner: str
    :param owner: ownerId
    :type env: str
    :param env: wiliot environment (prod/test/dev)
    :rtype: str
    :return: hearbeat table name
    """
    env = 'prod' if env is None else env
    hearbeat_table = f'_network_data_statistics_{env}'
    return hearbeat_table

def get_configuration_table_name(owner, env=None):
    """
    function gets ownerId and environment and return the name of the configuration table (in databricks)
    :type owner: str
    :param owner: ownerId
    :type env: str
    :param env: wiliot environment (prod/test/dev)
    :rtype: str
    :return: configuration table name
    """
    env = 'prod' if env is None else env
    configuration_table = owner + '' + f'_network_configuration_{env}'
    if owner.isnumeric():
        configuration_table = '_' + configuration_table
    return configuration_table

def process_graph(fig, graph_name, display_graph=True, save_graph=False, directory=None, pass_as_is=False):
    """
    function processes graph - displays and saves the graph (according to entered flags)
    :type fig: plotly Figure
    :param fig: graph to display
    :type graph_name: str
    :param graph_name: name of graph, and the filename by which it will be saved
    :type display_graph: bool
    :param display_graph: flag to choose if to display the graph in DataBricks Notebook
    :type save_graph: bool
    :param save_graph: flag to choose if to save the graph
    :type directory: str
    :param directory: directory to save graph
    :type pass_as_is: bool
    :param pass_as_is: if true the fig entered is already a plot
    """
    debug_print(f'Processing Graph - {graph_name}', center=True)
    if save_graph:
        if directory is None:
            exit_notebook('Need to supply directory to save graph!')
        if directory[-1] != '/':
            directory = directory + '/'
        filepath = f"""{directory}{graph_name}.html"""
        Path(directory).mkdir(parents=True, exist_ok=True)
        open(filepath, 'a').close()
        fig.write_html(filepath)
        create_download_link(filepath, file_name=graph_name)
    if display_graph:
        if is_databricks():
            if not pass_as_is:
                fig = plot(fig, output_type='div')
            display_html(fig)
        else:
            fig.show(renderer="browser")

def get_spark():
    try:
        from pyspark.sql import SparkSession
        return SparkSession.builder.getOrCreate()
    except NameError:
        raise NameError('Spark is not installed!')


def analyze_packets_data_per_brg(df, brg_list, time_start, test_time_conf_received=None):
    df = df.drop_duplicates(subset=["rawPacket", "bridgeId"])
    tag_stats_df = pd.DataFrame()
    for brg in brg_list:
        brg_df = df[df['bridgeId'] == brg]
        debug_print(f'Processing {brg_df.shape[0]} packets from {brg}')
        if brg_df.empty:
            continue
        df_tagstats = analyze_packets_data(packets_df=brg_df, test_time_start=time_start,
                                           test_time_conf_received=test_time_conf_received)
        tag_stats_dict = df_tagstats.get_group_statistics(group_by_col='externalId')
        tag_stats_df_tmp = pd.DataFrame.from_dict(tag_stats_dict, orient='index')
        tag_stats_df_tmp['bridgeId'] = brg
        tag_stats_df = pd.concat([tag_stats_df, tag_stats_df_tmp])
    if 'externalId' not in tag_stats_df.columns:
        tag_stats_df['externalId'] = tag_stats_df.index
    tag_stats_df = tag_stats_df.sort_values('rx_rate_normalized', ascending=False).drop_duplicates(
        'externalId').sort_index()
    return tag_stats_df


def flag_bugged_tags(rawdata_df):
    """
    function gets raw data dataframe, adds 'isBugged' column
    'isBugged' is true for each packet of a bugged tag, false otherwise
    :type rawdata_df: pandas DataFrame
    :param rawdata_df: raw packet data
    """
    rawdata_df['isBugged'] = None
    unique_tags = rawdata_df['externalId'].unique()
    for tag in unique_tags:
        tmp_df = rawdata_df.loc[rawdata_df['externalId'] == tag]
        tmp_df = tmp_df[['packet_counter', 'timestamp']]
        tmp_df = tmp_df.sort_values(by=['timestamp'])
        prev_packet_counter = 0
        cycles = 0
        orig_cycles = tmp_df['packet_counter'].unique()
        for timestamp in tmp_df['timestamp'].unique():
            # TODO - see when this function throws error
            try:
                packet_counter = tmp_df.loc[tmp_df['timestamp'] == timestamp, 'packet_counter'].unique().item()
            except Exception:
                continue
            if packet_counter + 256 * cycles < prev_packet_counter:
                cycles = cycles + 1
            prev_packet_counter = packet_counter + 256 * cycles
            tmp_df.loc[tmp_df['timestamp'] == timestamp, 'packet_counter'] = prev_packet_counter
        tmp_df2 = tmp_df.diff(axis=0)
        tmp_df2['rate'] = (1000 * tmp_df2['packet_counter'] / tmp_df2['timestamp'])
        max_rate = tmp_df2['rate'].max()
        is_bugged = False
        if max_rate > 6:
            is_bugged = True
            debug_print(f'{tag} is bugged! Flagging packets')
        rawdata_df.loc[rawdata_df['externalId'] == tag, 'isBugged'] = is_bugged


def parse_commastring(string):
    """
    parse string with comma or comma+space seperated values to list
    :type string: str
    :param string: input
    :rtype: list
    :return: list of values
    """
    if type(string) == float:
        try:
            if np.isnan(string):
                return None
        except Exception as e:
            pass
    if type(string) == list:
        return string
    if string is None:
        return list()
    cmd_list = ''.join(string.split()).split(',')
    return cmd_list


def parse_commastring_array(array):
    """
    parses Pandas array (DataFrame column) to list of all unique values
    :type array: Pandas array
    :param array: array
    :rtype: list
    :return: list of unique values
    """
    if len(array) == 1:
        return parse_commastring(array[0])
    result = list()
    for item in array:
        item_values = parse_commastring(item)
        if len(item_values) == 1:
            item_values
            if item_values[0] not in result:
                result.extend(item_values)
        else:
            for value in item_values:
                if value not in result:
                    result.extend([value])
    return result


import argparse
import numpy as np
import matplotlib.pyplot as plt
import mplcursors
import pandas as pd
import math
import pkg_resources
from enum import Enum
import sys

CSV_NAME = 'current_gain_data_set.csv'
PACKET_TABLE_CSV_PATH = pkg_resources.resource_filename(__name__, CSV_NAME)

class BridgeType(Enum):
    MINEW = {
        "avg_beacon_current": 208,
        "beacon_tx_time": 1.5,
        "avg_rx_current": 17.5,
        "max_output_power": 27.1
    }
    FANSTEL = {
        "avg_beacon_current": 101,
        "beacon_tx_time": 1.52,
        "avg_rx_current": 42.5,
        "max_output_power": 27.6
    }

class OptimalPowerConfiguration:
    def __init__(self, bridge_type, dc, rxtx_period):
        self.bridge_type = bridge_type
        self.dc_list = dc
        self.rxtx_period_list = rxtx_period
        self.avg_beacon_current = None
        self.beacon_tx_time = None
        self.avg_rx_current = None
        self.max_output_power = None
        self.data_df = None
        self.dc_rxtx_dict = None
        self.data_array = None
        self.dc_to_color = None

    def generate_current_gain_plot(self):
        self.load_csv_data_base()
        self.match_const_val_to_board_type()
        self.create_dc_rxtx_dict()
        self.calculate_data_create_np_array()
        self.create_color_mapping()
        self.plot_data()

    def load_csv_data_base(self):
        self.data_df = pd.read_csv(PACKET_TABLE_CSV_PATH)
        
    def match_const_val_to_board_type(self):
        try:
            bridge_type = BridgeType[self.bridge_type.upper()]  
            bridge_data = bridge_type.value
            self.avg_beacon_current = bridge_data["avg_beacon_current"]
            self.beacon_tx_time = bridge_data["beacon_tx_time"]
            self.avg_rx_current = bridge_data["avg_rx_current"]
            self.max_output_power = bridge_data["max_output_power"]
        except KeyError:
            print(f"Invalid bridge type: {self.bridge_type}. Please choose from the following bridge types: {', '.join(member.name.lower() for member in BridgeType)}.")
            sys.exit(1)

    def create_dc_rxtx_dict(self):
        self.dc_rxtx_dict = {}
        if len(self.dc_list) != len(self.rxtx_period_list):
            print("The number of 'dc' arguments must match the number of 'rxtx_perid' argument")
            sys.exit(1)
        for i, element in enumerate(self.dc_list):
            self.dc_rxtx_dict[element] = self.rxtx_period_list[i]

    def calculate_data_create_np_array(self):
        data = []
        self.data_df = self.data_df[self.data_df['bridge_type'] == self.bridge_type]
        for key in self.dc_rxtx_dict:
            for idx, row in self.data_df.iterrows():
                cfg_power = row['cfg_power']
                register_cfg = row['register_changed']
                avg_current = round(
                    key * row['X_current'] +
                    self.avg_beacon_current * (self.beacon_tx_time / self.dc_rxtx_dict[key]) +
                    self.avg_rx_current * ((self.dc_rxtx_dict[key] - self.beacon_tx_time) / self.dc_rxtx_dict[key]), 2)
                relative_gain = round((row['output_power'] - self.max_output_power) +
                                       10 * math.log10(key / 0.33), 2)
                pa = row['PA']
                data.append([key, cfg_power, avg_current, relative_gain, pa, register_cfg])
        self.data_array = np.array(data)

    def create_color_mapping(self):
        unique_dc_values = np.unique(self.data_array[:, 0])
        num_colors = len(unique_dc_values)
        custom_cmap = plt.get_cmap('viridis', num_colors)
        self.dc_to_color = {dc: custom_cmap(i) for i, dc in enumerate(unique_dc_values)}

    def plot_data(self):
        plt.figure(figsize=(12, 8))
        scatter = plt.scatter(self.data_array[:, 3], self.data_array[:, 2],
                            c=[self.dc_to_color[dc] for dc in self.data_array[:, 0]], marker='o', alpha=0.7)
        labels = ['DC: {:.2f}\nCfg Value: {}\nAvg Current: {:.2f}\nRelative Gain: {:.2f}\nPA: {}\nReg. changed: {}'.format(
            dc, cfg_power, avg_current, relative_gain, pa,register_cfg) for dc, cfg_power, avg_current, relative_gain, pa, register_cfg in self.data_array]

        mplcursors.cursor(scatter, hover=True).connect(
            "add", lambda sel: sel.annotation.set_text(labels[sel.index]))

        plt.xlabel('Relative Gain [dBm]')
        plt.ylabel('Avg Current [mA]')
        plt.title(f'Avg Current vs. Relative Gain for {self.bridge_type}')
        plt.grid()

        # Create a legend for the registers configurations
        legend_labels = [
        'Reg. cfg 1: pa_duty_cycle = 0x02; hp_max = 0x02;',
        'Reg. cfg 2: pa_duty_cycle = 0x02; hp_max = 0x03;',
        'Reg. cfg 3: pa_duty_cycle = 0x03; hp_max = 0x05;',
        'Reg. cfg 4: pa_duty_cycle = 0x04; hp_max = 0x07;']

        proxy_artists = [plt.Line2D([0], [0], marker='o', color='w', label=label,
                                    markerfacecolor='k', markersize=10) for label in legend_labels]

        plt.legend(handles=proxy_artists, title="Register configuration table:", loc='best')

        plt.show()

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Calculate and plot Avg Current vs. Relative Gain")
    required = parser.add_argument_group('required arguments')
    required.add_argument("-bridge_type", type=str, help="Bridge type (minew or fanstel)", required=True)
    required.add_argument("-dc", nargs="+", type=float, help="The values of the duty cycles to compare. The order of insertion must match the order of the corresponding rxtx period", required=True)
    required.add_argument("-rxtx_period", nargs="+", type=int, help="The value of rxtx period. The order of insertion must match the order of the corresponding duty cycle", required=True)
    args = parser.parse_args()

    opt_power_cfg = OptimalPowerConfiguration(args.bridge_type, args.dc, args.rxtx_period)
    opt_power_cfg.generate_current_gain_plot()
















